# vi: ts=4 sw=4

import logging
from typing import Any, Optional

import numpy as np

from .utils.epics_pvs import fmt_time

from .device import Component as Cpt
from .device import Device, required_for_connection
from .ophydobj import Kind
from .positioner import PositionerBase
from .signal import EpicsSignal, InternalSignal
from .status import wait as status_wait

logger = logging.getLogger(__name__)


class PVPositioner(Device, PositionerBase):
    """A Positioner which is controlled using multiple user-defined signals

    Keyword arguments are passed through to the base class, Positioner

    Parameters
    ----------
    prefix : str, optional
        The device prefix used for all sub-positioners. This is optional as it
        may be desirable to specify full PV names for PVPositioners.
    limits : 2-element sequence, optional
        (low_limit, high_limit)
    name : str
        The device name
    egu : str, optional
        The engineering units (EGU) for the position
    settle_time : float, optional
        The amount of time to wait after moves to report status completion
    timeout : float, optional
        The default timeout to use for motion requests, in seconds.

    Attributes
    ----------
    setpoint : Signal
        The setpoint (request) signal
    readback : Signal or None
        The readback PV (e.g., encoder position PV)
    actuate : Signal or None
        The actuation PV to set when movement is requested
    actuate_value : any, optional
        The actuation value, sent to the actuate signal when motion is
        requested
    stop_signal : Signal or None
        The stop PV to set when motion should be stopped
    stop_value : any, optional
        The value sent to stop_signal when a stop is requested
    done : Signal
        A readback value indicating whether motion is finished
    done_value : any, optional
        The value that the done pv should be when motion has completed
    put_complete : bool, optional
        If set, the specified PV should allow for asynchronous put completion
        to indicate motion has finished.  If `actuate` is specified, it will be
        used for put completion.  Otherwise, the `setpoint` will be used.  See
        the `-c` option from `caput` for more information.
    """

    setpoint = None  # TODO: should add limits=True
    readback = None
    actuate = None
    actuate_value = 1

    stop_signal = None
    stop_value = 1

    done = None
    done_value = 1
    put_complete = False

    def __init__(
        self,
        prefix="",
        *,
        limits=None,
        name=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        egu="",
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            name=name,
            parent=parent,
            **kwargs,
        )

        if self.__class__ is PVPositioner:
            raise TypeError(
                "PVPositioner must be subclassed with the correct "
                "signals set in the class definition."
            )

        self._egu = egu

        if limits is not None:
            self._limits = tuple(limits)
        else:
            self._limits = None

        if self.readback is not None:
            self.readback.subscribe(self._pos_changed)
            self.readback.kind = Kind.hinted
        elif self.setpoint is not None:
            self.setpoint.subscribe(self._pos_changed)
        else:
            raise ValueError("A setpoint or a readback must be specified")

        if self.done is None and not self.put_complete:
            msg = (
                'PVPositioner {} is mis-configured. A "done" Signal must be'
                " provided or use PVPositionerPC (which uses put completion"
                " to determine when motion has completed)."
                "".format(self.name)
            )
            raise ValueError(msg)

        if self.done is not None:
            self.done.subscribe(self._move_changed)
        else:
            # If there is not a `move_changed` signal, indicate that the
            # positioner is not moving frm the start:
            self._move_changed(value=self.done_value)

    @property
    def egu(self):
        """The engineering units (EGU) for a position"""
        return self._egu

    @property
    def put_complete(self):
        return isinstance(self, PVPositionerPC)

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        if self.limits is not None:
            low, high = self.limits
            if low != high and not (low <= pos <= high):
                raise ValueError("{} outside of user-specified limits" "".format(pos))
        else:
            self.setpoint.check_value(pos)

    @property
    def moving(self):
        """Whether or not the motor is moving

        If a `done` PV is specified, it will be read directly to get the motion
        status. If not, it determined from the internal state of PVPositioner.

        Returns
        -------
        bool
        """
        if self.done is not None:
            dval = self.done.get(use_monitor=False)
            return dval != self.done_value
        else:
            return self._moving

    def _setup_move(self, position):
        """Move and do not wait until motion is complete (asynchronous)"""
        self.log.debug("%s.setpoint = %s", self.name, position)
        self.setpoint.put(position, wait=True)
        if self.actuate is not None:
            self.log.debug("%s.actuate = %s", self.name, self.actuate_value)
            self.actuate.put(self.actuate_value, wait=False)

    def move(self, position, wait=True, timeout=None, moved_cb=None):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        # Before moving, ensure we can stop (if a stop_signal is configured).
        if self.stop_signal is not None:
            self.stop_signal.wait_for_connection()
        status = super().move(position, timeout=timeout, moved_cb=moved_cb)

        has_done = self.done is not None
        if not has_done:
            moving_val = 1 - self.done_value
            self._move_changed(value=self.done_value)
            self._move_changed(value=moving_val)

        try:
            self._setup_move(position)
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    @required_for_connection
    def _move_changed(self, timestamp=None, value=None, sub_type=None, **kwargs):
        was_moving = self._moving
        self._moving = value != self.done_value

        started = False
        if not self._started_moving:
            started = self._started_moving = not was_moving and self._moving
            self.log.debug(
                "[ts=%s] %s started moving: %s", fmt_time(timestamp), self.name, started
            )

        self.log.debug(
            "[ts=%s] %s moving: %s (value=%s)",
            fmt_time(timestamp),
            self.name,
            self._moving,
            value,
        )

        if started:
            self._run_subs(
                sub_type=self.SUB_START, timestamp=timestamp, value=value, **kwargs
            )

        if not self.put_complete:
            # In the case of put completion, motion complete
            if was_moving and not self._moving:
                self._done_moving(success=True, timestamp=timestamp, value=value)

    @required_for_connection
    def _pos_changed(self, timestamp=None, value=None, **kwargs):
        """Callback from EPICS, indicating a change in position"""
        self._set_position(value)

    def stop(self, *, success=False):
        if self.stop_signal is not None:
            self.stop_signal.put(self.stop_value, wait=False)
        super().stop(success=success)

    @property
    def report(self):
        rep = super().report
        rep["pv"] = self.readback.pvname
        return rep

    @property
    def limits(self):
        if self._limits is not None:
            return tuple(self._limits)
        else:
            return self.setpoint.limits

    def _repr_info(self):
        yield from super()._repr_info()

        yield ("limits", self._limits)
        yield ("egu", self._egu)

    def _done_moving(self, **kwargs):
        has_done = self.done is not None
        if not has_done:
            self._move_changed(value=self.done_value)

        super()._done_moving(**kwargs)


class PVPositionerPC(PVPositioner):
    def __init__(self, *args, **kwargs):
        if self.__class__ is PVPositionerPC:
            raise TypeError(
                "PVPositionerPC must be subclassed with the "
                "correct signals set in the class definition."
            )

        super().__init__(*args, **kwargs)

    def _setup_move(self, position):
        """Move and do not wait until motion is complete (asynchronous)"""

        def done_moving(**kwargs):
            self.log.debug("%s async motion done", self.name)
            self._done_moving(success=True)

        if self.done is None:
            # No done signal, so we rely on put completion
            moving_val = 1 - self.done_value
            self._move_changed(value=moving_val)

        self.log.debug("%s.setpoint = %s", self.name, position)

        if self.actuate is not None:
            self.setpoint.put(position, wait=True)

            self.log.debug("%s.actuate = %s", self.name, self.actuate_value)
            self.actuate.put(self.actuate_value, wait=False, callback=done_moving)
        else:
            self.setpoint.put(position, wait=False, callback=done_moving)


class PVPositionerComparator(PVPositioner):
    """
    PV Positioner with a software done signal.

    The done state is set by a comparison function defined in the class body.
    The comparison function takes two arguments, readback and setpoint,
    returning True if we are considered done or False if we are not.

    This class is intended to support `PVPositionerIsClose`, but exists to
    allow some flexibility if we want to use other metrics for deciding if
    the PVPositioner is done.

    Internally, this will subscribe to both the `setpoint` and `readback`
    signals, updating `done` as appropriate.

    Parameters
    ----------
    prefix : str, optional
        The device prefix used for all sub-positioners. This is optional as it
        may be desirable to specify full PV names for PVPositioners.
    limits : 2-element sequence, optional
        (low_limit, high_limit)
    name : str
        The device name
    egu : str, optional
        The engineering units (EGU) for the position
    settle_time : float, optional
        The amount of time to wait after moves to report status completion
    timeout : float, optional
        The default timeout to use for motion requests, in seconds.

    Attributes
    ----------
    setpoint : Signal
        The setpoint (request) signal
    readback : Signal or None
        The readback PV (e.g., encoder position PV)
    actuate : Signal or None
        The actuation PV to set when movement is requested
    actuate_value : any, optional
        The actuation value, sent to the actuate signal when motion is
        requested
    stop_signal : Signal or None
        The stop PV to set when motion should be stopped
    stop_value : any, optional
        The value sent to stop_signal when a stop is requested
    put_complete : bool, optional
        If set, the specified PV should allow for asynchronous put completion
        to indicate motion has finished.  If `actuate` is specified, it will be
        used for put completion.  Otherwise, the `setpoint` will be used.  See
        the `-c` option from `caput` for more information.
    """

    done = Cpt(InternalSignal, value=0)
    done_value = 1

    def __init__(self, prefix: str, *, name: str, **kwargs):
        self._last_readback = None
        self._last_setpoint = None
        super().__init__(prefix, name=name, **kwargs)
        if None in (self.setpoint, self.readback):
            raise NotImplementedError(
                "PVPositionerComparator requires both "
                "a setpoint and a readback signal to "
                "compare!"
            )

    def __init_subclass__(cls, **kwargs):
        """Set up callbacks in subclass."""
        super().__init_subclass__(**kwargs)
        if None not in (cls.setpoint, cls.readback):
            cls.setpoint.sub_value(cls._update_setpoint)
            cls.readback.sub_value(cls._update_readback)

    def done_comparator(self, readback: Any, setpoint: Any) -> bool:
        """
        Override done_comparator in your subclass.

        This method should return True if we are done moving
        and False otherwise.
        """
        raise NotImplementedError("Must implement a done comparator!")

    def _update_setpoint(self, *args, value: Any, **kwargs) -> None:
        """Callback to cache the setpoint and update done state."""
        self._last_setpoint = value
        # Always set done to False when a move is requested
        # This means we always get a rising edge when finished moving
        # Even if the move distance is under our done moving tolerance
        self.done.put(0, internal=True)
        self._update_done()

    def _update_readback(self, *args, value: Any, **kwargs) -> None:
        """Callback to cache the readback and update done state."""
        self._last_readback = value
        self._update_done()

    def _update_done(self) -> None:
        """Update our status to done if we pass the comparator."""
        if None not in (self._last_readback, self._last_setpoint):
            is_done = self.done_comparator(self._last_readback, self._last_setpoint)
            done_value = int(is_done)
            if done_value != self.done.get():
                self.done.put(done_value, internal=True)


class PVPositionerIsClose(PVPositionerComparator):
    """
    PV Positioner that updates done state based on np.isclose.

    Effectively, this will treat our move as complete if the readback is
    sufficiently close to the setpoint. This is generically helpful for
    PV positioners that don't have a `done` signal built into the hardware.

    The arguments atol and rtol can be set as class attributes or passed as
    initialization arguments.

    atol is a measure of absolute tolerance. If atol is 0.1, then you'd be
    able to be up to 0.1 units away and still count as done. This is
    typically the most useful parameter for calibrating done tolerance.

    rtol is a measure of relative tolerance. If rtol is 0.1, then you'd be
    able to deviate from the goal position by up to 10% of its value. This
    is useful for small quantities. For example, defining an atol for a
    positioner that ranges from 1e-8 to 2e-8 could be somewhat awkward.

    Parameters
    ----------
    prefix : str, optional
        The device prefix used for all sub-positioners. This is optional as it
        may be desirable to specify full PV names for PVPositioners.
    limits : 2-element sequence, optional
        (low_limit, high_limit)
    name : str
        The device name
    egu : str, optional
        The engineering units (EGU) for the position
    settle_time : float, optional
        The amount of time to wait after moves to report status completion
    timeout : float, optional
        The default timeout to use for motion requests, in seconds.
    atol : float, optional
        A measure of absolute tolerance. If atol is 0.1, then you'd be
        able to be up to 0.1 units away and still count as done.
    rtol : float, optional
        A measure of relative tolerance. If rtol is 0.1, then you'd be
        able to deviate from the goal position by up to 10% of its value

    Attributes
    ----------
    setpoint : Signal
        The setpoint (request) signal
    readback : Signal or None
        The readback PV (e.g., encoder position PV)
    actuate : Signal or None
        The actuation PV to set when movement is requested
    actuate_value : any, optional
        The actuation value, sent to the actuate signal when motion is
        requested
    stop_signal : Signal or None
        The stop PV to set when motion should be stopped
    stop_value : any, optional
        The value sent to stop_signal when a stop is requested
    put_complete : bool, optional
        If set, the specified PV should allow for asynchronous put completion
        to indicate motion has finished.  If `actuate` is specified, it will be
        used for put completion.  Otherwise, the `setpoint` will be used.  See
        the `-c` option from `caput` for more information.
    atol : float, optional
        A measure of absolute tolerance. If atol is 0.1, then you'd be
        able to be up to 0.1 units away and still count as done.
    rtol : float, optional
        A measure of relative tolerance. If rtol is 0.1, then you'd be
        able to deviate from the goal position by up to 10% of its value
    """

    atol: Optional[float] = None
    rtol: Optional[float] = None

    def __init__(
        self,
        prefix: str,
        *,
        name: str,
        atol: Optional[float] = None,
        rtol: Optional[float] = None,
        **kwargs,
    ):
        if atol is not None:
            self.atol = atol
        if rtol is not None:
            self.rtol = rtol
        super().__init__(prefix, name=name, **kwargs)

    def done_comparator(self, readback: float, setpoint: float) -> bool:
        """
        Check if the readback is close to the setpoint value.

        Uses numpy.isclose to make the comparison. Tolerance values
        atol and rtol for numpy.isclose are taken from the attributes
        self.atol and self.rtol, which can be defined as class attributes
        or passed in as init parameters.

        If atol or rtol are omitted, the default values from numpy are
        used instead.
        """
        kwargs = {}
        if self.atol is not None:
            kwargs["atol"] = self.atol
        if self.rtol is not None:
            kwargs["rtol"] = self.rtol
        return np.isclose(readback, setpoint, **kwargs)


class PVPositionerDone(PVPositioner):
    """
    PV Positioner with no readback that reports done immediately.

    This is for the case where you'd like a PV to look like a
    positioner, but the truth is that it is just a PV.

    When the user asks for a move, set done to 0 and then back to 1.
    This simulates the normal positioner behavior.

    Parameters
    ----------
    prefix : str, optional
        The device prefix used for all sub-positioners. This is optional as it
        may be desirable to specify full PV names for PVPositioners.
    limits : 2-element sequence, optional
        (low_limit, high_limit)
    name : str
        The device name
    egu : str, optional
        The engineering units (EGU) for the position
    settle_time : float, optional
        The amount of time to wait after moves to report status completion
    timeout : float, optional
        The default timeout to use for motion requests, in seconds.

    Attributes
    ----------
    actuate : Signal or None
        The actuation PV to set when movement is requested
    actuate_value : any, optional
        The actuation value, sent to the actuate signal when motion is
        requested
    stop_signal : Signal or None
        The stop PV to set when motion should be stopped
    stop_value : any, optional
        The value sent to stop_signal when a stop is requested
    put_complete : bool, optional
        If set, the specified PV should allow for asynchronous put completion
        to indicate motion has finished.  If `actuate` is specified, it will be
        used for put completion.  Otherwise, the `setpoint` will be used.  See
        the `-c` option from `caput` for more information.
    """

    setpoint = Cpt(EpicsSignal, "", kind="hinted")

    done = Cpt(InternalSignal, value=0)
    done_value = 1

    def _setup_move(self, position: Any) -> None:
        super()._setup_move(position)
        self.done.put(0, internal=True)
        self.done.put(1, internal=True)



from ophyd import (
    Device,
    Signal,
    Kind,
    Component,
    ALL_COMPONENTS,
    kind_context,
    DynamicDeviceComponent as DDCpt,
)
import pytest


def test_standalone_signals():
    # A object's kind only matters when we read its _parent_. It affects
    # whether its parent recursively calls read() and/or read_configuration().
    # When we call read() and/or read_configuration() on the object itself,
    # directly, and its 'kind' setting does not affects its behavior.
    normal_sig = Signal(name="normal_sig", value=3, kind=Kind.normal)
    assert normal_sig.kind == Kind.normal
    assert normal_sig.name not in normal_sig.hints["fields"]
    assert "normal_sig" in normal_sig.read()
    assert "normal_sig" in normal_sig.read_configuration()

    hinted_sig = Signal(name="hinted_sig", value=3)
    assert hinted_sig.kind == Kind.hinted
    assert hinted_sig.name in hinted_sig.hints["fields"]
    assert "hinted_sig" in hinted_sig.read()
    assert "hinted_sig" in hinted_sig.read_configuration()

    # Same with a sig set up this way
    config_sig = Signal(name="config_sig", value=5, kind=Kind.config)
    assert config_sig.kind == Kind.config
    assert config_sig.name not in config_sig.hints["fields"]
    assert "config_sig" in config_sig.read()
    assert "config_sig" in config_sig.read_configuration()

    # Same with a sig set up this way
    omitted_sig = Signal(name="omitted_sig", value=5, kind=Kind.omitted)
    assert omitted_sig.kind == Kind.omitted
    assert omitted_sig.name not in omitted_sig.hints["fields"]
    assert "omitted_sig" in omitted_sig.read()
    assert "omitted_sig" in omitted_sig.read_configuration()

    # But these will be differentiated when we put them into a parent
    # Device and call read() or read_configuraiton() on _that_.


def test_nested_devices():
    class A(Device):
        normal_sig = Component(Signal)
        config_sig = Component(Signal, kind=Kind.config)
        omitted_sig = Component(Signal, kind=Kind.omitted)

    a = A(name="a")

    # When we call read and read_configuration on a, it checks the kind of its
    # components and reads the right ones.
    assert "a_normal_sig" in a.read()
    assert "a_config_sig" not in a.read()
    assert "a_omitted_sig" not in a.read()

    assert "a_normal_sig" not in a.read_configuration()
    assert "a_config_sig" in a.read_configuration()
    assert "a_omitted_sig" not in a.read_configuration()

    # Another layer of nesting!

    class B(Device):
        a_default = Component(A)
        a_config = Component(A, kind=Kind.config)
        a_omitted = Component(A, kind=Kind.omitted)

    b = B(name="b")

    assert ["b_a_default_normal_sig"] == list(b.read())

    # Notice that a_default comes along for the ride here. If you ask
    # a Device for its normal readings it will also give you its
    # configuration. (You need complete configurational metadata for
    # the EventDescriptor!)
    assert ["b_a_default_config_sig", "b_a_config_config_sig"] == list(
        b.read_configuration()
    )
    # Notice that it tacks 'config' on when you set the kind to 'normal'.
    assert (Kind.normal | Kind.config) == B(name="b", kind=Kind.normal).kind
    # And the same if you try to set it after __init__
    b.a_default.kind = Kind.normal
    assert b.a_default.kind == (Kind.normal | Kind.config)
    # However, just taking 'config' alone without Event-wise readings is fine.
    b.a_default.kind = Kind.config
    assert b.a_default.kind == Kind.config
    # Now we get no Event-wise readings from a_default
    assert [] == list(b.read())
    assert ["b_a_default_config_sig", "b_a_config_config_sig"] == list(
        b.read_configuration()
    )


def test_strings():
    sig = Signal(name="sig", kind="normal")
    assert sig.kind == Kind.normal


def test_kind_context():
    class Thing(Device):

        with kind_context("omitted") as Cpt:
            a = Cpt(Signal)

    thing = Thing(name="thing")
    assert thing.a.kind == Kind.omitted


# Test back-compatibility. The expected values in these tests match the
# behaviors from before the Kind feature was introduced --- modulo the _order_.


def test_behavior_if_nothing_is_specified():
    class Thing(Device):
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name="t")
    assert set("ab") == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)


def test_class_default_attrs_both_none():

    # Test various values of _default_read_attrs and
    # _default_configuration_attrs

    class Thing(Device):
        _default_read_attrs = None
        _default_configuration_attrs = None
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name="t")
    assert set("ab") == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)
    assert ["t_a", "t_b"] == list(t.describe())
    assert set() == set(t.describe_configuration())

    class ThingHaver(Device):
        _default_read_attrs = None
        _default_configuration_attrs = None
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name="th")
    assert set(["a", "b"] + ["a.a", "a.b", "b.a", "b.b"]) == set(th.read_attrs)

    # THIS IS ACTUALLY _NOT_ BACKWARD-COMPATIBLE BEHAVIOR!
    # But we consider the former behavior a bug. Unlike a Signal, a Device is
    # not allowed to place itself in its parent's read_attrs ONLY. It must also
    # be in configuration_attrs if it is in read_attrs. (Under the hood, this
    # is enforced in the setter of the Device's `kind` property, which OR's the
    # kind value with Kind.config if said vlaue includes Kind.normal.
    assert set("ab") == set(th.configuration_attrs)

    assert ["th_a_a", "th_a_b", "th_b_a", "th_b_b"] == list(th.describe())
    assert [] == list(t.describe_configuration())


def test_default_read_attrs_empty_and_configuration_attrs_none():
    class Thing(Device):
        _default_read_attrs = []
        _default_configuration_attrs = None
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name="t")
    assert set() == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)

    class ThingHaver(Device):
        _default_read_attrs = []
        _default_configuration_attrs = None
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name="th")
    assert set() == set(th.read_attrs)
    assert [] == list(th.describe())
    assert set() == set(th.configuration_attrs)
    assert [] == list(th.describe_configuration())


def test_default_read_attrs_none_and_configuration_attrs_empty():
    class Thing(Device):
        _default_read_attrs = None
        _default_configuration_attrs = []
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name="t")
    assert set("ab") == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)

    class ThingHaver(Device):
        _default_read_attrs = None
        _default_configuration_attrs = []
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name="th")
    assert set(["a", "b"] + ["a.a", "a.b", "b.a", "b.b"]) == set(th.read_attrs)
    assert ["th_a_a", "th_a_b", "th_b_a", "th_b_b"] == list(th.describe())
    assert set("ab") == set(th.configuration_attrs)
    assert [] == list(th.describe_configuration())


def test_default_attrs_both_empty():
    class Thing(Device):
        _default_read_attrs = []
        _default_configuration_attrs = []
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name="t")
    assert set() == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)

    class ThingHaver(Device):
        _default_read_attrs = []
        _default_configuration_attrs = []
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name="th")
    assert set() == set(th.read_attrs)
    assert [] == list(th.describe())
    assert set() == set(th.configuration_attrs)
    assert [] == list(t.describe_configuration())


def test_all_components_escape_hatch():
    class Thing(Device):
        _default_read_attrs = ["a"]
        _default_configuration_attrs = ["b"]
        a = Component(Signal)
        b = Component(Signal)

    class ThingEscapeHatch(Thing):
        _default_read_attrs = ALL_COMPONENTS
        _default_configuration_attrs = []
        c = Component(Signal)

    t = ThingEscapeHatch(name="t")
    assert set("abc") == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)


def test_default_attrs_nonempty_disjoint():
    class Thing(Device):
        _default_read_attrs = ["a"]
        _default_configuration_attrs = ["b"]
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name="t")
    assert set("a") == set(t.read_attrs)
    assert set("b") == set(t.configuration_attrs)

    class ThingHaver(Device):
        _default_read_attrs = ["a"]
        _default_configuration_attrs = ["b"]
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name="th")
    assert set(["a", "a.a"]) == set(th.read_attrs)
    assert "b" not in th.read_attrs
    assert ["th_a_a"] == list(th.describe())
    # Here again we have the same break from back-compatability. We are not
    # allowed to put a sub-Device in configuration_attrs that is not also in
    # read_attrs. The old code would have:
    # assert set(['b']) == set(th.configuration_attrs)
    # But now we get:
    assert set(["a", "b"] + ["a.b", "b.b"]) == set(th.configuration_attrs)
    assert ["th_a_b", "th_b_b"] == list(th.describe_configuration())


def test_options_via_init():
    ...


def test_back_compat():
    # Test class defaults.
    class Thing(Device):
        _default_read_attrs = ["a"]
        _default_configuration_attrs = ["b"]
        a = Component(Signal)
        b = Component(Signal)
        c = Component(Signal)

    thing = Thing(name="thing")
    assert ["thing_a"] == list(thing.read())
    # assert ['thing_b'] == list(thing.read_configuration())

    # Test attribute getting and setting.
    assert thing.read_attrs == ["a"]
    assert thing.configuration_attrs == ["b"]
    thing.read_attrs = ["a", "b"]
    assert ["thing_a", "thing_b"] == list(thing.read())

    thing = Thing(name="thing", read_attrs=["b"], configuration_attrs=["a"])
    assert ["thing_b"] == list(thing.read())


@pytest.fixture(scope="function")
def thing_haver_haver():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal)
        d = Component(Signal, kind=Kind.hinted)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted)
        B = Component(Thing, kind=Kind.config)
        C = Component(Thing, kind=Kind.normal)

    class ThingHaverHaver(Device):
        alpha = Component(ThingHaver, kind=Kind.omitted)
        beta = Component(ThingHaver, kind=Kind.config)
        gamma = Component(ThingHaver, kind=Kind.normal)

    return ThingHaverHaver(name="thh")


def test_list_proxy(thing_haver_haver):

    thh = thing_haver_haver
    assert "gamma" in thh.read_attrs
    assert "gamma" in list(thh.read_attrs)

    assert "beta" in thh.configuration_attrs
    assert "beta" in list(thh.configuration_attrs)

    assert "alpha" not in thh.read_attrs
    assert "alpha" not in thh.configuration_attrs

    assert "alpha" not in list(thh.read_attrs)
    assert "alpha" not in list(thh.configuration_attrs)

    assert "gamma.C" in thh.read_attrs
    assert "gamma.C" in list(thh.read_attrs)

    assert "gamma.C.c" in list(thh.read_attrs)
    assert "gamma.C.d" in list(thh.read_attrs)


def test_ddc():
    class Thing(Device):
        a = Component(Signal)
        b = Component(Signal)

    class ThingHaver(Device):
        a = DDCpt({"A": (Thing, "", {})})

    th = ThingHaver(name="th")
    assert th.a.A.a.kind & Kind.normal
    assert th.a.A.kind & Kind.normal
    assert th.a.kind & Kind.normal
    assert th.kind & Kind.normal


def test_no_lazy():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal)
        d = Component(Signal, kind=Kind.hinted)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted)
        B = Component(Thing, kind=Kind.config)
        C = Component(Thing, kind=Kind.normal)

    th = ThingHaver(name="th")
    assert set("ABC") == set(th._signals)
    for k in "ABC":
        assert k in th._signals
        assert set("abcd") == set(getattr(th, k)._signals)


def test_some_lazy():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal, lazy=True)
        d = Component(Signal, kind=Kind.hinted, lazy=True)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted, lazy=True)
        B = Component(Thing, kind=Kind.config)
        C = Component(Thing, kind=Kind.normal)

    th = ThingHaver(name="th")
    assert set("BC") == set(th._signals)
    for k in "BC":
        c = getattr(th, k)
        assert set("ab") == set(c._signals)


def test_all_lazy():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal, lazy=True)
        d = Component(Signal, kind=Kind.hinted, lazy=True)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted, lazy=True)
        B = Component(Thing, kind=Kind.config, lazy=True)
        C = Component(Thing, kind=Kind.normal, lazy=True)

    th = ThingHaver(name="th")
    assert not th._signals


def test_grand_child_breaks_lazy():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal, lazy=True)
        d = Component(Signal, kind=Kind.hinted, lazy=True)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted, lazy=True)
        B = Component(Thing, kind=Kind.config, lazy=True)
        C = Component(Thing, kind=Kind.normal, lazy=True)

    th = ThingHaver(name="th", read_attrs=["A.a", "A.d"])
    assert set("A") == set(th._signals)
    A = th.A
    assert set("ab") == set(A._signals)



#!/usr/bin/env python3
from caproto.server import pvproperty, PVGroup, ioc_arg_parser, run


class SignalTestIOC(PVGroup):
    read_only = pvproperty(value=0.0, read_only=True, alarm_group="alarm_a")
    read_write = pvproperty(
        value=0.0,
        lower_ctrl_limit=-100.0,
        upper_ctrl_limit=100.0,
        alarm_group="alarm_a",
    )

    pair_rbv = pvproperty(value=0.0, read_only=True, alarm_group="alarm_a")
    pair_set = pvproperty(
        value=0.0,
        lower_ctrl_limit=-100.0,
        upper_ctrl_limit=100.0,
        alarm_group="alarm_a",
    )

    @pair_set.putter
    async def pair_set(self, instance, value):
        await self.pair_rbv.write(value=value)

    waveform = pvproperty(
        value=[ord("a"), ord("b"), ord("c")], read_only=True, alarm_group="alarm_a"
    )
    bool_enum = pvproperty(value=True, alarm_group="alarm_a")
    alarm_status = pvproperty(value=0)
    set_severity = pvproperty(value=0)

    @set_severity.putter
    async def set_severity(self, instance, severity):
        await self.read_only.alarm.write(
            severity=severity, status=self.alarm_status.value
        )

    INITIAL_PATH = "/path/here"
    path = pvproperty(value=INITIAL_PATH, max_length=255)
    path_RBV = pvproperty(value=INITIAL_PATH, max_length=255)

    @path.putter
    async def path(self, instance, value):
        await self.path_RBV.write(value=value)
        return value


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="signal_tests:", desc="ophyd.tests.test_signal test IOC"
    )
    ioc = SignalTestIOC(**ioc_options)
    run(ioc.pvdb, **run_options)



from collections import OrderedDict

from . import (
    EpicsSignalRO,
    EpicsSignal,
    Component as Cpt,
    DynamicDeviceComponent as DDCpt,
    Signal,
    Kind,
    kind_context,
)
from .areadetector import (
    EpicsSignalWithRBV,
    ImagePlugin,
    StatsPlugin,
    DetectorBase,
    SingleTrigger,
    ADBase,
)
from .status import DeviceStatus


def _current_fields(attr_base, field_base, range_, **kwargs):
    defn = OrderedDict()
    for i in range_:
        attr = "{attr}{i}".format(attr=attr_base, i=i)
        suffix = "{field}{i}".format(field=field_base, i=i)
        defn[attr] = (EpicsSignal, suffix, kwargs)

    return defn


class QuadEMPort(ADBase):
    port_name = Cpt(Signal, value="")

    def __init__(self, port_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.port_name.put(port_name)


class QuadEM(SingleTrigger, DetectorBase):
    # These settings intentionally shadow the settings inherited from
    # DetectorBase.
    _default_read_attrs = None
    _default_configuration_attrs = None

    _status_type = DeviceStatus  # overrriding the default in SingleTrigger

    # This is needed because ophyd verifies that it can see all
    # of the nodes in the asyn pipeline, however these IOCs do not
    # expose their port name via a PV, but nevertheless server as the
    # root node for the plugins.
    # Leaving this port_name here for compatibility
    integration_time = Cpt(EpicsSignalWithRBV, "IntegrationTime", kind="config")
    averaging_time = Cpt(EpicsSignalWithRBV, "AveragingTime", kind="config")
    with kind_context("omitted") as OCpt:
        conf = Cpt(QuadEMPort, port_name="NSLS_EM")
        model = OCpt(EpicsSignalRO, "Model")
        firmware = OCpt(EpicsSignalRO, "Firmware")

        acquire_mode = OCpt(EpicsSignalWithRBV, "AcquireMode")
        acquire = OCpt(EpicsSignal, "Acquire")

        read_format = OCpt(EpicsSignalWithRBV, "ReadFormat")
        em_range = OCpt(EpicsSignalWithRBV, "Range")
        ping_pong = OCpt(EpicsSignalWithRBV, "PingPong")

        num_channels = OCpt(EpicsSignalWithRBV, "NumChannels")
        geometry = OCpt(EpicsSignalWithRBV, "Geometry")
        resolution = OCpt(EpicsSignalWithRBV, "Resolution")

        bias_state = OCpt(EpicsSignalWithRBV, "BiasState")
        bias_interlock = OCpt(EpicsSignalWithRBV, "BiasInterlock")
        bias_voltage = OCpt(EpicsSignalWithRBV, "BiasVoltage")
        hvs_readback = OCpt(EpicsSignalRO, "HVSReadback")
        hvv_readback = OCpt(EpicsSignalRO, "HVVReadback")
        hvi_readback = OCpt(EpicsSignalRO, "HVIReadback")

        values_per_read = OCpt(EpicsSignalWithRBV, "ValuesPerRead")
        sample_time = OCpt(EpicsSignalRO, "SampleTime_RBV")  # yay consistency
        num_average = OCpt(EpicsSignalRO, "NumAverage_RBV")
        num_averaged = OCpt(EpicsSignalRO, "NumAveraged_RBV")
        num_acquire = OCpt(EpicsSignalWithRBV, "NumAcquire")
        num_acquired = OCpt(EpicsSignalRO, "NumAcquired")
        read_data = OCpt(EpicsSignalRO, "ReadData")
        ring_overflows = OCpt(EpicsSignalRO, "RingOverflows")
        trigger_mode = OCpt(EpicsSignal, "TriggerMode")
        reset = OCpt(EpicsSignal, "Reset")

    current_names = DDCpt(
        _current_fields("ch", "CurrentName", range(1, 5), string=True)
    )
    current_offsets = DDCpt(_current_fields("ch", "CurrentOffset", range(1, 5)))
    current_offset_calcs = DDCpt(
        _current_fields("ch", "ComputeCurrentOffset", range(1, 5))
    )
    current_scales = DDCpt(_current_fields("ch", "CurrentScale", range(1, 5)))

    position_offset_x = Cpt(EpicsSignal, "PositionOffsetX")
    position_offset_y = Cpt(EpicsSignal, "PositionOffsetY")

    position_offset_calc_x = Cpt(EpicsSignal, "ComputePosOffsetX")
    position_offset_calc_y = Cpt(EpicsSignal, "ComputePosOffsetY")

    position_scale_x = Cpt(EpicsSignal, "PositionScaleX")
    position_scale_Y = Cpt(EpicsSignal, "PositionScaleY")

    image = Cpt(ImagePlugin, "image1:")
    current1 = Cpt(StatsPlugin, "Current1:")
    current2 = Cpt(StatsPlugin, "Current2:")
    current3 = Cpt(StatsPlugin, "Current3:")
    current4 = Cpt(StatsPlugin, "Current4:")

    sum_all = Cpt(StatsPlugin, "SumAll:")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.stage_sigs.update(
            [("acquire", 0), ("acquire_mode", 2)]  # if acquiring, stop  # single mode
        )
        self._acquisition_signal = self.acquire

        for i in range(1, 5):
            current = getattr(self, "current{}".format(i))
            current.mean_value.kind = Kind.hinted


class NSLS_EM(QuadEM):
    ...


class TetrAMM(QuadEM):
    conf = Cpt(QuadEMPort, port_name="TetrAMM", kind="omitted")


class APS_EM(QuadEM):
    conf = Cpt(QuadEMPort, port_name="APS_EM", kind="omitted")



import io
import logging
import logging.handlers

import ophyd.log as log
from ophyd.ophydobj import OphydObject
from ophyd.status import Status


def test_validate_level():
    log.validate_level("CRITICAL")
    log.validate_level("ERROR")
    log.validate_level("WARNING")
    log.validate_level("INFO")
    log.validate_level("DEBUG")
    log.validate_level("NOTSET")


def test_default_config_ophyd_logging():
    log.config_ophyd_logging()

    assert isinstance(log.current_handler, logging.StreamHandler)
    assert log.logger.getEffectiveLevel() <= logging.WARNING
    assert log.control_layer_logger.getEffectiveLevel() <= logging.WARNING


def test_config_ophyd_logging():
    datefmt = "%Y:%m:%d %H-%M-%S"

    log.config_ophyd_logging(
        file="ophyd.log",
        datefmt=datefmt,
        color=False,
        level="DEBUG",
    )

    assert isinstance(log.current_handler, logging.FileHandler)
    assert log.current_handler.formatter.datefmt == datefmt
    assert log.logger.getEffectiveLevel() <= logging.DEBUG
    assert log.control_layer_logger.getEffectiveLevel() <= logging.DEBUG


def test_logger_adapter_ophyd_object():
    log_buffer = io.StringIO()
    log_stream = logging.StreamHandler(stream=log_buffer)
    log_stream.setFormatter(log.LogFormatter())

    log.logger.addHandler(log_stream)

    ophyd_object = OphydObject(name="testing OphydObject.log")
    ophyd_object.log.info("here is some info")
    assert log_buffer.getvalue().endswith(
        "[testing OphydObject.log] here is some info\n"
    )


def test_logger_adapter_status():
    log_buffer = io.StringIO()
    log_stream = logging.StreamHandler(stream=log_buffer)
    log_stream.setFormatter(log.LogFormatter())

    log.logger.addHandler(log_stream)

    status = Status()
    status.log.info("here is some info")
    assert log_buffer.getvalue().endswith(f"[{str(status)}] here is some info\n")

    status.set_finished()
    status.log.info("here is more info")
    assert log_buffer.getvalue().endswith(f"[{str(status)}] here is more info\n")



import functools
import inspect
import re
import sys
import textwrap
from collections import OrderedDict
from typing import (
    Any,
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

import networkx as nx
import numpy as np

from ..device import Component, Device, DynamicDeviceComponent
from ..ophydobj import Kind, OphydObject
from ..signal import ArrayAttributeSignal, DerivedSignal, EpicsSignal, EpicsSignalRO
from . import docs


class EpicsSignalWithRBV(EpicsSignal):
    # An EPICS signal that simply uses the areaDetector convention of
    # 'pvname' being the setpoint and 'pvname_RBV' being the read-back

    def __init__(self, prefix, **kwargs):
        super().__init__(prefix + "_RBV", write_pv=prefix, **kwargs)


class NDDerivedSignal(DerivedSignal):
    """
    DerivedSignal to shape a flattened array

    The purpose of this class is to take a flattened array and shape in its
    proper form. The shape of the final array may be static in which case the
    shape and number of dimensions can be set as static integers. Otherwise,
     other signals from this classes parent can inform the proper shape of
    the array.

    Parameters
    ----------
    derived_from : str, ``ophyd.Signal``
        Either a `str` that specifies the attribute of the parent we will get
        the unshaped array or an ``ophyd.Signal`` itself.

    shape: tuple
        A tuple containing integers in the case of a static array or links to
        ``Signal`` objects. The specifications of the signals follows the same
        rules as the ``derived_from`` parameter

    num_dimensions: int, str, ``Signal``, optional
        The number of dimensions of the array. This can either be a static
        ``int`` or a ``Signal`` itself.

    Example
    -------
    .. code:: python

        class TwoDimensionalDetector(Device):

            flat_array = Cpt(EpicsSignal, ':Array')
            width = Cpt(EpicsSignalRO, ':Width')
            height = Cpt(EpicsSignalRO, ':Height')
            shaped_array = Cpt(NDDerivedSignal, 'flat_array',
                                               shape=('height', 'width'),
                                               num_dimensions=2)
    """

    def __init__(
        self, derived_from, *, shape, num_dimensions=None, parent=None, **kwargs
    ):
        # Assemble our shape of signals
        self._shape = []
        self._has_subscribed = False
        for dim in shape:
            if isinstance(dim, str):
                dim = getattr(parent, dim)
            self._shape.append(dim)
        self._shape = tuple(self._shape)
        # Assemble ndims
        if not num_dimensions:
            num_dimensions = len(self._shape)
        if isinstance(num_dimensions, str):
            num_dimensions = getattr(parent, num_dimensions)
        self._num_dimensions = num_dimensions
        super().__init__(derived_from=derived_from, parent=parent, **kwargs)

    @property
    def derived_shape(self):
        """Shape of output signal"""
        shape = list()
        for dim in self._shape:
            if not isinstance(dim, (int, float)):
                dim = dim.get()
            shape.append(dim)
        return tuple(shape)

    @property
    def derived_ndims(self):
        """Number of dimensions"""
        ndims = self._num_dimensions
        if not isinstance(ndims, (int, float)):
            ndims = ndims.get()
        return int(ndims)

    def forward(self, value):
        """Flatten the array to send back to the DerivedSignal"""
        return np.array(value).flatten()

    def inverse(self, value):
        """Shape the flat array to send as a result of ``.get``"""
        array_shape = self.derived_shape[: self.derived_ndims]
        if not any(array_shape):
            raise RuntimeError(f"Invalid array size {self.derived_shape}")

        array_len = np.prod(array_shape)
        if len(value) < array_len:
            raise RuntimeError(
                f"cannot reshape array of size {len(value)} "
                f"into shape {tuple(array_shape)}. Check IOC configuration."
            )

        return np.asarray(value[:array_len]).reshape(array_shape)

    def subscribe(self, callback, event_type=None, run=True):
        cid = super().subscribe(callback, event_type=event_type, run=run)
        if not self._has_subscribed and (
            event_type is None or event_type == self.SUB_VALUE
        ):
            # Ensure callbacks are fired when array is reshaped
            for dim in self._shape + (self._num_dimensions,):
                if not isinstance(dim, int):
                    dim.subscribe(
                        self._array_shape_callback, event_type=self.SUB_VALUE, run=False
                    )
        self._has_subscribed = True
        return cid

    def _array_shape_callback(self, **kwargs):
        # TODO we need a better way to say "latest new is good enough"
        value = self.inverse(self._derived_from._readback)
        self._readback = value
        self._run_subs(sub_type=self.SUB_VALUE, value=value, **self._metadata)


K = TypeVar("K", bound=OphydObject)


class ADComponent(Component[K]):
    #: Default laziness for the component.  All AreaDetector components are
    #: by default lazy - as they contain thousands of PVs that aren't
    #: necesssary for everyday functionality.
    lazy_default: ClassVar[bool] = True

    #: The attribute name of the component.
    attr: Optional[str]
    #: The class to instantiate when the device is created.
    cls: Type[K]
    #: Keyword arguments for the device creation.
    kwargs: Dict[str, Any]
    #: Lazily create components on access.
    lazy: bool
    #: PV or identifier suffix.
    suffix: Optional[str]
    #: Documentation string.
    doc: Optional[str]
    #: Value to send on ``trigger()``
    trigger_value: Optional[Any]
    #: The data acquisition kind.
    kind: Kind
    #: Names of kwarg keys to prepend the device PV prefix to.
    add_prefix: Tuple[str, ...]
    #: Subscription name -> subscriptions marked by decorator.
    _subscriptions: DefaultDict[str, List[Callable]]

    def find_docs(self, parent_class):
        """Find all the documentation related to this class, all the way up the
        MRO"""

        classes = inspect.getmro(parent_class)
        for class_ in classes:
            try:
                html_file = class_._html_docs
            except AttributeError:
                continue

            for fn in html_file:
                if fn in docs.docs:
                    yield docs.docs[fn]

    def make_docstring(self, parent_class):
        """Create a docstring for the component, given the parent class"""

        def make_codeblock(s):
            """Make a codeblock that will render nicely in sphinx"""
            block = [
                "AreaDetector Component",
                "::",
                "",
            ]

            lines = s.split("\n", 1)
            header, lines = lines[0], lines[1:]

            block.append(textwrap.indent(textwrap.dedent(header), prefix=" " * 4))

            lines = "\n".join(lines)
            block.append(textwrap.indent(textwrap.dedent(lines), prefix=" " * 4))
            block.append("")
            return "\n".join(block)

        if self.suffix is None:
            return

        suffixes = [self.suffix]

        if self.suffix.endswith("_RBV"):
            suffixes.append(self.suffix[:-4])

        for doc in self.find_docs(parent_class):
            for suffix in suffixes:
                try:
                    return make_codeblock(doc[suffix])
                except KeyError:
                    pass

        return super().make_docstring(parent_class)


def ad_group(cls, attr_suffix, **kwargs):
    """Definition creation for groups of signals in areadetectors"""
    defn = OrderedDict()
    kwargs.setdefault("lazy", True)
    for attr, suffix in attr_suffix:
        defn[attr] = (cls, suffix, kwargs)
    return defn


def _ddc_helper(signal_class, *items, kind="config", doc=None, **kwargs):
    "DynamicDeviceComponent using one signal class for all components"
    return DynamicDeviceComponent(
        ad_group(signal_class, items, kind=kind, **kwargs),
        doc=doc,
    )


DDC_EpicsSignal = functools.partial(_ddc_helper, EpicsSignal)
DDC_EpicsSignalRO = functools.partial(_ddc_helper, EpicsSignalRO)
DDC_SignalWithRBV = functools.partial(_ddc_helper, EpicsSignalWithRBV)


class ADBase(Device):
    """The AreaDetector base class

    This serves as the base for all detectors and plugins
    """

    _html_docs = ["areaDetectorDoc.html"]
    _default_read_attrs = ()
    _default_configuration_attrs = ()

    def find_signal(
        self, text, use_re=False, case_sensitive=False, match_fcn=None, f=sys.stdout
    ):
        """Search through the signal docs on this detector for the string text

        Parameters
        ----------
        text : str
            Text to find
        use_re : bool, optional
            Use regular expressions
        case_sensitive : bool, optional
            Case sensitive search
        match_fcn : callable, optional
            Function to call when matches are found Defaults to a function that
            prints matches to f
        f : file-like, optional
            File-like object that the default match function prints to
            (Defaults to sys.stdout)
        """
        # TODO: Some docstrings change based on the detector type,
        #       showing different options than are available in
        #       the base area detector class (for example). As such,
        #       instead of using the current docstrings, this grabs
        #       them again.

        def default_match(attr, signal, doc):
            print("Property: {}".format(attr), file=f)
            print("  Signal: {!r}".format(signal), file=f)
            print("     Doc: {}".format(doc), file=f)
            print(file=f)

        if match_fcn is None:
            match_fcn = default_match

        if use_re:
            flags = re.MULTILINE
            if not case_sensitive:
                flags |= re.IGNORECASE

            regex = re.compile(text, flags=flags)

        elif not case_sensitive:
            text = text.lower()

        for attr, cpt in self._sig_attrs.items():
            doc = cpt.make_docstring(self.__class__)

            match = False
            if use_re:
                if regex.search(doc):
                    match = True
            else:
                if not case_sensitive:
                    if text in doc.lower():
                        match = True
                elif text in doc:
                    match = True

            if match:
                match_fcn(attr=attr, signal=getattr(self, attr), doc=doc)

    def stage(self, *args, **kwargs):
        ret = super().stage(*args, **kwargs)
        try:
            self.validate_asyn_ports()
        except RuntimeError as err:
            self.unstage(*args, **kwargs)
            raise err
        return ret

    def get_plugin_by_asyn_port(self, port_name):
        """Get the plugin which has the given asyn port name

        Parameters
        ----------
        port_name : str
            The port name to search for

        Returns
        -------
        ret : ADBase or None
            Either the requested plugin or None if not found

        """
        try:
            name = self.port_name.get()
        except AttributeError:
            pass
        else:
            if name == port_name:
                return self

        for name, subdevice in self.walk_subdevices(include_lazy=True):
            if hasattr(subdevice, "get_plugin_by_asyn_port"):
                sig = subdevice.get_plugin_by_asyn_port(port_name)
                if sig is not None:
                    return sig
        return None

    def get_asyn_port_dictionary(self):
        """Return port name : component map

        Returns
        -------
        port_map : dict
            Mapping between port_name and ADBase objects
        """
        # uniqueness of port names enforced at IOC layer
        ret = {}
        try:
            ret.update({self.port_name.get(): self})
        except AttributeError:
            pass

        for name, subdevice in self.walk_subdevices(include_lazy=True):
            if hasattr(subdevice, "get_asyn_port_dictionary"):
                ret.update(subdevice.get_asyn_port_dictionary())

        return ret

    def get_asyn_digraph(self):
        """Get the directed graph of the ASYN ports

        Returns
        -------
        G : networkx.DiGraph
            Directed graph of pipelines

        port_map : dict
            Mapping between port_name and ADBase objects
        """
        port_map = self.get_asyn_port_dictionary()
        G = nx.DiGraph()
        for out_port, cpt in port_map.items():
            try:
                in_port = cpt.nd_array_port.get()
            except AttributeError:
                # attribute error because we hit a component which is not
                # a plugin, but is the 'base' data source
                G.add_node(out_port)
            else:
                G.add_edge(in_port, out_port)

        return G, port_map

    def visualize_asyn_digraph(self, ax=None, *args, **kwargs):
        """This generates a figure showing the current asyn port layout.

        This method generates a plot showing all of the currently enabled
        Areadetector plugin asyn ports and their relationships. The current
        ports and relationships are found using self.get_asyn_digraph.

        Parameters
        ----------
        ax: matplotlib axes
            if None (default) then a new figure is created otherwise it is
            plotted on the specified axes.
        *args, **kwargs : networkx.draw_networkx args and kwargs.
            For the allowed args and kwargs see the `networkx.draw_networkx documentation
            <https://networkx.github.io/documentation/networkx-1.10/reference/generated/networkx.drawing.nx_pylab.draw_networkx.html>`_
        """
        # Importing matplotlib.pyplot here as it is not a dependency except for
        # this method.
        import matplotlib.pyplot as plt

        # Generate the port_map Digraph.
        G, port_map = self.get_asyn_digraph()
        # Create and label the figure if no ax is provided.
        if not ax:
            fig, ax = plt.subplots()
            ax.set_title("AD port map for {}".format(self.name))
            plt.tick_params(
                axis="x", which="both", bottom=False, top=False, labelbottom=False
            )
            plt.tick_params(
                axis="y", which="both", left=False, right=False, labelbottom=False
            )

        nx.draw_networkx(G, ax=ax, *args, **kwargs)

    def validate_asyn_ports(self):
        """Validate that all components of pipeline are known

        Raises
        ------
        RuntimeError
           If there any input ports to known plugins where the source is
           not known to ophyd
        """
        g, port_map = self.get_asyn_digraph()
        g = nx.Graph(g)
        if port_map and nx.number_connected_components(g) != 1:
            missing_plugins = self.missing_plugins()
            raise RuntimeError(
                "The asyn ports {!r} are used by plugins "
                "that ophyd is aware of but the source plugin "
                "is not.  Please reconfigure your device to "
                "include the source plugin or reconfigure "
                "to not use these ports."
                "".format(missing_plugins)
            )

    def missing_plugins(self):
        """Find missing ports"""
        g, port_map = self.get_asyn_digraph()
        ret = []

        for node in g:
            if node not in port_map:
                ret.append(node)

        return ret

    configuration_names = Component(
        ArrayAttributeSignal, attr="_configuration_names", kind="config"
    )

    @property
    def _configuration_names(self):
        return [getattr(self, c).name for c in self.configuration_attrs]

    @property
    def ad_root(self):
        root = self
        while True:
            if not isinstance(root.parent, ADBase):
                return root
            root = root.parent



# vi: ts=4 sw=4 sts=4 expandtab
import inspect
import logging
from collections import OrderedDict
import warnings

from .errors import *  # noqa: F401, F403
from .epics_pvs import *  # noqa: F401, F403
from .paths import makedirs, make_dir_tree  # noqa: F401

logger = logging.getLogger(__name__)


def enum(**enums):
    """Create an enum from the keyword arguments"""
    return type("Enum", (object,), enums)


class OrderedDefaultDict(OrderedDict):
    """
    a combination of defaultdict and OrderedDict

    source: http://stackoverflow.com/a/6190500/1221924
    """

    def __init__(self, default_factory=None, *a, **kw):
        if default_factory is not None and not callable(default_factory):
            raise TypeError("first argument must be callable")
        super().__init__(*a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = (self.default_factory,)
        return type(self), args, None, None, self.items()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        import copy

        return type(self)(self.default_factory, copy.deepcopy(self.items()))

    def __repr__(self):
        return "%s(%s, %s)" % (
            self.__class__.__name__,
            self.default_factory,
            super().__repr__(),
        )


def doc_annotation_forwarder(base_klass):
    def wrapper(f):
        f_name = getattr(f, "__name__")
        base_func = getattr(base_klass, f_name)
        base_docs = getattr(base_func, "__doc__")
        base_annotation = getattr(base_func, "__annotations__")
        f.__doc__ = base_docs
        f.__annotations__ = base_annotation

        return f

    return wrapper


def _filtered_ip_ns():
    import IPython

    return {
        k: v for k, v in IPython.get_ipython().user_ns.items() if not k.startswith("_")
    }


def instances_from_namespace(classes, *, ns=None):
    """Get instances of `classes` from the user namespace

    Parameters
    ----------
    classes : type, or sequence of types
        Passed directly to isinstance(), only instances of these classes
        will be returned.

    ns : Dict[str, Any], optional
       namespace to pull from, defaults to getting the
    """
    if ns is None:
        ns = _filtered_ip_ns()
    return [val for val in ns.values() if isinstance(val, classes)]


def ducks_from_namespace(attrs, *, ns=None):
    """Get instances that have all of attributes.

    "Ducks" is a reference to "duck-typing." If it looks like a duck....

    Parameters
    ----------
    attr : Union[str, Iterable[str]]
        name of attribute or list of names
    """
    if isinstance(attrs, str):
        attrs = [attrs]
    if ns is None:
        ns = _filtered_ip_ns()
    return [val for val in ns.values() if all(hasattr(val, attr) for attr in attrs)]


def underscores_to_camel_case(underscores):
    "Convert abc_def_ghi to abcDefGhi"
    if "_" in underscores and underscores.strip("_") != "":
        return "".join(s.capitalize() for s in underscores.split("_"))

    return underscores.capitalize()


def getattrs(obj, gen):
    "For each item in the generator `gen`, yields (attr, getattr(obj, attr))"
    for attr in gen:
        yield attr, getattr(obj, attr)


class DO_NOT_USE:
    "sentinel value"
    ...


def adapt_old_callback_signature(callback):
    """
    If callback has signature callback(), wrap in signature callback(status).

    Parameters
    ----------
    callback: callable
        Expected signature ``callback(status)`` or ``callback()``

    Returns
    -------
    callback: callable
        Signature ``callback(status)``
    """
    # Handle callback with signature callback() for back-compat.
    sig = inspect.signature(callback)
    try:
        # Does this callback accept one positional argument?
        sig.bind(None)
    except TypeError:
        warnings.warn(
            "The signature of a Status callback is now expected to "
            "be cb(status). The signature cb() is "
            "supported, but support will be removed in a future release "
            "of ophyd.",
            DeprecationWarning,
            stacklevel=3,
        )
        raw_callback = callback

        def callback(status):
            # Do nothing with status because the user-provided callback cannot
            # accept it.
            raw_callback()

    return callback



import os
import calendar


def makedirs(path, *, mode=0o777, mode_base=None):
    """Recursively make directories and set permissions

    Parameters
    ----------
    path : str
        Full path to create, including all parent directories
    mode : int, optional
        Mode to set
    mode_base : str, optional
        If specified, only try to set mode after this directory
        Full base path
    """
    # Permissions not working with os.makedirs -
    # See: http://stackoverflow.com/questions/5231901
    if not path or os.path.exists(path):
        return []

    head, tail = os.path.split(path)

    # Recurse on all directories above:
    ret = makedirs(head, mode=mode, mode_base=mode_base)

    # Then try to make the last directory:
    os.makedirs(path, mode=mode, exist_ok=True)

    # And set its permissions if after mode_base:
    if not mode_base or os.path.commonprefix((mode_base, path)) == mode_base:
        os.chmod(path, mode)

    ret.append(path)
    return ret


def make_dir_tree(year, *, base_path=None, mode=0o777):
    """Make full directory tree for the year

    Parameters
    ----------
    year : int
    base_path : str, optional
        If unspecified, defaults to the current directory
    mode : int, optional
        File mode to set for permissions (default: 0o777)

    Returns
    -------
    paths : list
        List of directories created
    """
    if base_path is None:
        base_path = os.getcwd()

    year_dir = os.path.join(base_path, str(year))
    paths = []

    for month in range(1, 13):
        month_dir = os.path.join(year_dir, "%02d" % month)

        _, num_days = calendar.monthrange(year, month)
        for day in range(1, 1 + num_days):
            day_path = os.path.join(month_dir, "%02d" % day)
            paths.append(day_path)

    for path in paths:
        makedirs(path, mode=mode, mode_base=base_path)

    return paths



from .. import Device, Component as Cpt
from . import plugins
from ..device import create_device_from_components
from ..ophydobj import select_version, OphydObject


class CommonPlugins(Device, version_type="ADCore"):
    ...


class CommonOverlayPlugin(Device, version_type="ADCore"):
    ...


class CommonAttributePlugin(Device, version_type="ADCore"):
    ...


class CommonROIStatPlugin(Device, version_type="ADCore"):
    ...


class CommonGatherPlugin(Device, version_type="ADCore"):
    ...


def _get_bases(cls, version):
    mixin_cls = select_version(cls, version)
    base_cls = select_version(plugins.PluginBase, version)
    if issubclass(mixin_cls, base_cls):
        return (mixin_cls,)

    return (mixin_cls, base_cls)


def _make_common_numbered(clsname, version, cpt_cls_base, bases, attr_prefix):
    try:
        cpt_cls = select_version(cpt_cls_base, version)
    except ValueError:
        return

    base_cls_base, plugin_base = bases
    bases = (base_cls_base,) + _get_bases(plugin_base, version)

    return create_device_from_components(
        name=clsname,
        base_class=bases,
        class_kwargs={"version_of": base_cls_base, "version": version},
        **{f"{attr_prefix}_{j}": Cpt(cpt_cls, f"{j}:") for j in range(1, 9)},
    )


def _make_common_gather(clsname, version):
    base_cls_base = plugins.GatherPlugin
    cpt_cls_base = plugins.GatherNPlugin

    try:
        cpt_cls = select_version(cpt_cls_base, version)
    except ValueError:
        return

    bases = (CommonGatherPlugin,) + _get_bases(base_cls_base, version)

    return create_device_from_components(
        name=clsname,
        base_class=bases,
        class_kwargs={"version_of": CommonGatherPlugin, "version": version},
        **{f"gather_{j}": Cpt(cpt_cls, "", index=j) for j in range(1, 9)},
    )


versions = [
    (1, 9, 1),
    (2, 0),
    (2, 1),
    (2, 2),
    (2, 3),
    (2, 4),
    (2, 5),
    (2, 6),
    (3, 1),
    (3, 2),
    (3, 3),
    (3, 4),
]


for version in versions:
    ver_string = "".join(str(_) for _ in version)
    _make_common_numbered(
        f"CommonOverlayPlugin_V{ver_string}",
        version,
        plugins.Overlay,
        (
            CommonOverlayPlugin,
            plugins.OverlayPlugin,
        ),
        "overlay",
    )
    _make_common_numbered(
        f"CommonAttributePlugin_V{ver_string}",
        version,
        plugins.AttributeNPlugin,
        (CommonAttributePlugin, plugins.AttributePlugin),
        "attr",
    )
    _make_common_numbered(
        f"CommonROIStatPlugin_V{ver_string}",
        version,
        plugins.ROIStatNPlugin,
        (CommonROIStatPlugin, plugins.ROIStatPlugin),
        "attr",
    )

    _make_common_gather(f"CommonGatherPlugin_V{ver_string}", version)


all_plugins = [
    ("attr1", CommonAttributePlugin, "Attr1:"),
    ("cb1", plugins.CircularBuffPlugin, "CB1:"),
    ("cc1", plugins.ColorConvPlugin, "CC1:"),
    ("cc2", plugins.ColorConvPlugin, "CC2:"),
    ("codec1", plugins.CodecPlugin, "Codec1:"),
    ("codec2", plugins.CodecPlugin, "Codec2:"),
    ("fft1", plugins.FFTPlugin, "FFT1:"),
    ("gather1", CommonGatherPlugin, "Gather1:"),
    ("hdf1", plugins.HDF5Plugin, "HDF1:"),
    ("jpeg1", plugins.JPEGPlugin, "JPEG1:"),
    ("kafka1", plugins.KafkaPlugin, "KAFKA1:"),
    ("magick1", plugins.MagickPlugin, "Magick1:"),
    ("netcdf1", plugins.NetCDFPlugin, "netCDF1:"),
    ("nexus1", plugins.NexusPlugin, "Nexus1:"),
    ("over1", CommonOverlayPlugin, "Over1:"),
    ("proc1", plugins.ProcessPlugin, "Proc1:"),
    ("proc1_tiff", plugins.TIFFPlugin, "Proc1:TIFF:"),
    ("roi1", plugins.ROIPlugin, "ROI1:"),
    ("roi2", plugins.ROIPlugin, "ROI2:"),
    ("roi3", plugins.ROIPlugin, "ROI3:"),
    ("roi4", plugins.ROIPlugin, "ROI4:"),
    ("roistat1", CommonROIStatPlugin, "ROIStat1:"),
    ("scatter1", plugins.ScatterPlugin, "Scatter1:"),
    ("stats1", plugins.StatsPlugin, "Stats1:"),
    ("stats1_ts", plugins.TimeSeriesPlugin, "Stats1:TS:"),
    ("stats2", plugins.StatsPlugin, "Stats2:"),
    ("stats2_ts", plugins.TimeSeriesPlugin, "Stats2:TS:"),
    ("stats3", plugins.StatsPlugin, "Stats3:"),
    ("stats3_ts", plugins.TimeSeriesPlugin, "Stats3:TS:"),
    ("stats4", plugins.StatsPlugin, "Stats4:"),
    ("stats4_ts", plugins.TimeSeriesPlugin, "Stats4:TS:"),
    ("stats5", plugins.StatsPlugin, "Stats5:"),
    ("stats5_ts", plugins.TimeSeriesPlugin, "Stats5:TS:"),
    ("tiff1", plugins.TIFFPlugin, "TIFF1:"),
    ("trans1", plugins.TransformPlugin, "Trans1:"),
]


common_plugins = {}


for version in versions:
    local_plugins = {}
    for attr, cls, suffix in all_plugins:
        if attr == "magick1" and version > (3, 0):
            continue
        elif attr.endswith("_ts") and version < (3, 3):
            continue
        elif attr == "fft1" and version < (3, 0):
            continue
        elif attr == "proc1_tiff" and version < (3, 3):
            continue
        elif attr == "attr1" and version < (2, 2):
            continue

        try:
            cls = select_version(cls, version)
        except ValueError:
            continue

        local_plugins[attr] = Cpt(cls, suffix)

    ver_string = "".join(str(_) for _ in version)
    class_name = f"CommonPlugins_V{ver_string}"

    common_plugins[class_name] = create_device_from_components(
        name=class_name,
        base_class=CommonPlugins,
        class_kwargs=dict(version=version, version_of=CommonPlugins),
        **local_plugins,
    )


versioned_plugins = {}


class PluginNamespace(OphydObject, version_type="ADCore"):
    """
    This is intended to be used with select_versions.

    This provides namespaces of plugins for a given version of Area Detector.

    Examples
    --------
    Access the StatsPlugin for Area Detector version 3.4.

    >>> select_version(PluginNamespace, (3, 4)).StatsPlugin
    ophyd.areadetector.plugins.StatsPlugin_V34
    """

    # This is implemented as an OphydObject in order to leverage OphydObject's
    # version-resolution class keyword arguments. But it should *not* be used
    # as a Device or Signal. It should be used with select_version as in the
    # example in the docstring.
    ...


for version in versions:
    local_plugins = {}
    for _, cls, _ in all_plugins:
        try:
            local_plugins[cls.__name__] = select_version(cls, version)
        except ValueError:
            continue
    ver_string = "".join(str(_) for _ in version)
    class_name = f"PluginNamespace_V{ver_string}"
    versioned_plugins[class_name] = type(
        class_name,
        (PluginNamespace,),
        local_plugins,
        version=version,
        version_of=PluginNamespace,
    )


globals().update(**common_plugins, **versioned_plugins)
del local_plugins
del version
del class_name
del suffix
del ver_string
del attr
del cls


__all__ = ["PluginNamespace", "CommonPlugins"]



import logging
import time
import copy
import pytest
import threading

from ophyd import get_cl
from ophyd.signal import (
    Signal,
    EpicsSignal,
    EpicsSignalRO,
    DerivedSignal,
    InternalSignal,
    InternalSignalError,
    EpicsSignalNoValidation,
)
from ophyd.utils import ReadOnlyError, AlarmStatus, AlarmSeverity
from ophyd.status import wait
from ophyd.areadetector.paths import EpicsPathSignal

logger = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def ro_signal(cleanup, signal_test_ioc):
    sig = EpicsSignalRO(signal_test_ioc.pvs["pair_rbv"], name="pair_rbv")
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope="function")
def nv_signal(cleanup, signal_test_ioc):
    sig = EpicsSignalNoValidation(signal_test_ioc.pvs["pair_set"], name="pair_set")
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope="function")
def bool_enum_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(signal_test_ioc.pvs["bool_enum"], name="bool_enum")
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope="function")
def rw_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(signal_test_ioc.pvs["read_write"], name="read_write")
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope="function")
def pair_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(
        read_pv=signal_test_ioc.pvs["pair_rbv"],
        write_pv=signal_test_ioc.pvs["pair_set"],
        name="pair",
    )
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope="function")
def motor_pair_signal(cleanup, motor):
    sig = EpicsSignal(
        write_pv=motor.user_setpoint.pvname, read_pv=motor.user_readback.pvname
    )
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope="function")
def set_severity_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(signal_test_ioc.pvs["set_severity"], name="set_severity")
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope="function")
def alarm_status_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(signal_test_ioc.pvs["alarm_status"], name="alarm_status")
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


def test_signal_base():
    start_t = time.time()

    name = "test"
    value = 10.0
    signal = Signal(name=name, value=value, timestamp=start_t)
    signal.wait_for_connection()

    assert signal.connected
    assert signal.name == name
    with pytest.warns(UserWarning):
        assert signal.value == value
    assert signal.get() == value
    assert signal.timestamp == start_t

    info = dict(called=False)

    def _sub_test(**kwargs):
        info["called"] = True
        info["kw"] = kwargs

    signal.subscribe(_sub_test, run=False, event_type=signal.SUB_VALUE)
    assert not info["called"]

    signal.value = value
    signal.clear_sub(_sub_test)

    signal.subscribe(_sub_test, run=False, event_type=signal.SUB_VALUE)
    signal.clear_sub(_sub_test, event_type=signal.SUB_VALUE)

    kw = info["kw"]
    assert "value" in kw
    assert "timestamp" in kw
    assert "old_value" in kw

    assert kw["value"] == value
    assert kw["old_value"] == value
    assert kw["timestamp"] == signal.timestamp

    # readback callback for soft signal
    info = dict(called=False)
    signal.subscribe(_sub_test, event_type=Signal.SUB_VALUE, run=False)
    assert not info["called"]
    signal.put(value + 1)
    assert info["called"]

    signal.clear_sub(_sub_test)
    kw = info["kw"]

    assert "value" in kw
    assert "timestamp" in kw
    assert "old_value" in kw

    assert kw["value"] == value + 1
    assert kw["old_value"] == value
    assert kw["timestamp"] == signal.timestamp

    signal.trigger()
    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))


def test_signal_copy():
    start_t = time.time()

    name = "test"
    value = 10.0
    signal = Signal(name=name, value=value, timestamp=start_t)
    sig_copy = copy.copy(signal)

    assert signal.name == sig_copy.name
    with pytest.warns(UserWarning):
        assert signal.value == sig_copy.value
    assert signal.get() == sig_copy.get()
    assert signal.timestamp == sig_copy.timestamp


def test_signal_describe_fail():
    """
    Test Signal.describe() exception handling in the
    case where a Signal's value is not bluesky-friendly.
    """
    signal = Signal(name="the_none_signal", value=None)
    with pytest.raises(ValueError) as excinfo:
        signal.describe()
    assert "failed to describe 'the_none_signal' with value 'None'" in str(
        excinfo.value
    )


def test_internalsignal_write_from_internal():
    test_signal = InternalSignal(name="test_signal")
    for value in range(10):
        test_signal.put(value, internal=True)
        assert test_signal.get() == value
    for value in range(10):
        test_signal.set(value, internal=True).wait()
        assert test_signal.get() == value


def test_internalsignal_write_protection():
    test_signal = InternalSignal(name="test_signal")
    for value in range(10):
        with pytest.raises(InternalSignalError):
            test_signal.put(value)
        with pytest.raises(InternalSignalError):
            test_signal.set(value)


def test_epicssignal_readonly(cleanup, signal_test_ioc):
    signal = EpicsSignalRO(signal_test_ioc.pvs["read_only"])
    cleanup.add(signal)
    signal.wait_for_connection()
    print("EpicsSignalRO.metadata=", signal.metadata)
    signal.get()

    assert not signal.write_access
    assert signal.read_access

    with pytest.raises(ReadOnlyError):
        signal.value = 10

    with pytest.raises(ReadOnlyError):
        signal.put(10)

    with pytest.raises(ReadOnlyError):
        signal.set(10)

    # vestigial, to be removed
    with pytest.raises(AttributeError):
        signal.setpoint_ts

    # vestigial, to be removed
    with pytest.raises(AttributeError):
        signal.setpoint

    signal.precision
    signal.timestamp
    signal.limits

    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))
    time.sleep(0.2)


def test_epicssignal_novalidation(nv_signal):
    print("EpicsSignalNoValidation.metadata=", nv_signal.metadata)

    nv_signal.put(10)
    st = nv_signal.set(11)

    assert st.done

    nv_signal.get()
    nv_signal.read()

    nv_signal.describe()
    nv_signal.describe_configuration()

    nv_signal.read_configuration()


def test_epicssignal_readwrite_limits(pair_signal):
    signal = pair_signal
    signal.use_limits = True
    signal.check_value((signal.low_limit + signal.high_limit) / 2)

    with pytest.raises(ValueError):
        signal.check_value(None)

    with pytest.raises(ValueError):
        signal.check_value(signal.low_limit - 1)

    with pytest.raises(ValueError):
        signal.check_value(signal.high_limit + 1)


def test_epicssignal_readwrite(signal_test_ioc, pair_signal):
    pair_signal.use_limits = True
    signal = pair_signal

    assert signal.setpoint_pvname == signal_test_ioc.pvs["pair_set"]
    assert signal.pvname == signal_test_ioc.pvs["pair_rbv"]
    signal.get()

    time.sleep(0.2)

    value = 10
    signal.value = value
    signal.put(value)
    assert signal.setpoint == value
    signal.setpoint_ts

    signal.limits
    signal.precision
    signal.timestamp

    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))
    time.sleep(0.2)


def test_epicssignal_waveform(cleanup, signal_test_ioc):
    called = False

    def update_cb(value=None, **kwargs):
        nonlocal called
        assert len(value) > 1
        called = True

    signal = EpicsSignal(signal_test_ioc.pvs["waveform"], string=True)
    cleanup.add(signal)
    signal.wait_for_connection()

    sub = signal.subscribe(update_cb, event_type=signal.SUB_VALUE)
    assert len(signal.get()) > 1
    # force the current thread to allow other threads to run to service
    # subscription
    time.sleep(0.2)
    assert called
    signal.unsubscribe(sub)


def test_no_connection(cleanup, signal_test_ioc):
    sig = EpicsSignal("does_not_connect")
    cleanup.add(sig)

    with pytest.raises(TimeoutError):
        sig.wait_for_connection()

    sig = EpicsSignal("does_not_connect")
    cleanup.add(sig)

    with pytest.raises(TimeoutError):
        sig.put(0.0)

    with pytest.raises(TimeoutError):
        sig.get()

    sig = EpicsSignal(signal_test_ioc.pvs["read_only"], write_pv="does_not_connect")
    cleanup.add(sig)
    with pytest.raises(TimeoutError):
        sig.wait_for_connection()


def test_enum_strs(bool_enum_signal):
    assert bool_enum_signal.enum_strs == ("Off", "On")


def test_setpoint(rw_signal):
    rw_signal.get_setpoint()
    rw_signal.get_setpoint(as_string=True)


def test_epicssignalro():
    with pytest.raises(TypeError):
        # not in initializer parameters anymore
        EpicsSignalRO("test", write_pv="nope_sorry")


def test_describe(bool_enum_signal):
    sig = bool_enum_signal

    sig.put(1)
    desc = sig.describe()["bool_enum"]
    assert desc["dtype"] == "integer"
    assert desc["shape"] == []
    # assert 'precision' in desc
    assert desc["enum_strs"] == ("Off", "On")
    assert "upper_ctrl_limit" in desc
    assert "lower_ctrl_limit" in desc

    sig = Signal(name="my_pv")
    sig.put("Off")
    desc = sig.describe()["my_pv"]
    assert desc["dtype"] == "string"
    assert desc["shape"] == []

    sig.put(3.14)
    desc = sig.describe()["my_pv"]
    assert desc["dtype"] == "number"
    assert desc["shape"] == []

    import numpy as np

    sig.put(
        np.array(
            [
                1,
            ]
        )
    )
    desc = sig.describe()["my_pv"]
    assert desc["dtype"] == "array"
    assert desc["shape"] == [
        1,
    ]


def test_set_method():
    sig = Signal(name="sig")

    st = sig.set(28)
    wait(st)
    assert st.done
    assert st.success
    assert sig.get() == 28


def test_soft_derived():
    timestamp = 1.0
    value = "q"
    original = Signal(name="original", timestamp=timestamp, value=value)

    cb_values = []

    def callback(value=None, **kwargs):
        cb_values.append(value)

    derived = DerivedSignal(derived_from=original, name="derived")
    derived.subscribe(callback, event_type=derived.SUB_VALUE)

    assert derived.timestamp == timestamp
    assert derived.get() == value
    assert derived.timestamp == timestamp
    assert derived.describe()[derived.name]["derived_from"] == original.name
    assert derived.write_access == original.write_access
    assert derived.read_access == original.read_access

    new_value = "r"
    derived.put(new_value)
    assert original.get() == new_value
    assert derived.get() == new_value
    assert derived.timestamp == original.timestamp
    assert derived.limits == original.limits

    copied = copy.copy(derived)
    with pytest.warns(UserWarning):
        assert copied.derived_from.value == original.value
    assert copied.derived_from.timestamp == original.timestamp
    assert copied.derived_from.name == original.name

    derived.put("s")
    assert cb_values == ["r", "s"]

    called = []

    event = threading.Event()

    def meta_callback(*, connected, read_access, write_access, **kw):
        called.append(("meta", connected, read_access, write_access))
        event.set()

    derived.subscribe(meta_callback, event_type=derived.SUB_META, run=False)

    original._metadata["write_access"] = False
    original._run_subs(sub_type="meta", **original._metadata)

    event.wait(1)

    assert called == [("meta", True, True, False)]


def test_epics_signal_derived_ro(ro_signal):
    assert ro_signal.connected
    assert ro_signal.read_access
    assert not ro_signal.write_access

    derived = DerivedSignal(derived_from=ro_signal, name="derived")
    derived.wait_for_connection()

    assert derived.connected
    assert derived.read_access
    assert not derived.write_access

    assert derived.timestamp == ro_signal.timestamp
    assert derived.get() == ro_signal.get()


def test_epics_signal_derived_nv(nv_signal):
    assert nv_signal.connected
    assert nv_signal.read_access
    assert nv_signal.write_access

    derived = DerivedSignal(derived_from=nv_signal, name="derived")
    derived.wait_for_connection()

    assert derived.connected

    assert derived.timestamp == nv_signal.timestamp


@pytest.mark.motorsim
@pytest.mark.parametrize("put_complete", [True, False])
def test_epicssignal_set(motor_pair_signal, put_complete):
    sim_pv = motor_pair_signal
    sim_pv.put_complete = put_complete

    logging.getLogger("ophyd.signal").setLevel(logging.DEBUG)
    logging.getLogger("ophyd.utils.epics_pvs").setLevel(logging.DEBUG)
    print("tolerance=", sim_pv.tolerance)
    assert sim_pv.tolerance is not None

    start_pos = sim_pv.get()

    # move to +0.2 and check the status object
    target = sim_pv.get() + 0.2
    st = sim_pv.set(target, timeout=1, settle_time=0.001)
    wait(st, timeout=5)
    assert st.done
    assert st.success
    print("status 1", st)
    assert abs(target - sim_pv.get()) < 0.05

    # move back to -0.2, forcing a timeout with a low value
    target = sim_pv.get() - 0.2
    st = sim_pv.set(target, timeout=1e-6)
    time.sleep(0.5)
    print("status 2", st)
    assert st.done
    assert not st.success

    # keep the axis in position
    st = sim_pv.set(start_pos)
    wait(st, timeout=5)


statuses_and_severities = [
    (AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM),
    (AlarmStatus.READ, AlarmSeverity.MINOR),
    (AlarmStatus.WRITE, AlarmSeverity.MAJOR),
    (AlarmStatus.HIHI, AlarmSeverity.INVALID),
    (AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM),
]


@pytest.mark.parametrize("status, severity", statuses_and_severities)
def test_epicssignal_alarm_status(
    set_severity_signal, alarm_status_signal, pair_signal, status, severity
):
    alarm_status_signal.put(status, wait=True)
    set_severity_signal.put(severity, wait=True)

    pair_signal.get()
    assert pair_signal.alarm_status == status
    assert pair_signal.alarm_severity == severity

    pair_signal.get_setpoint()
    assert pair_signal.setpoint_alarm_status == status
    assert pair_signal.setpoint_alarm_severity == severity


@pytest.mark.parametrize("status, severity", statuses_and_severities)
def test_epicssignalro_alarm_status(
    set_severity_signal, alarm_status_signal, ro_signal, status, severity
):
    alarm_status_signal.put(status, wait=True)
    set_severity_signal.put(severity, wait=True)

    ro_signal.get()
    assert ro_signal.alarm_status == status
    assert ro_signal.alarm_severity == severity


def test_hints(cleanup, fake_motor_ioc):
    sig = EpicsSignalRO(fake_motor_ioc.pvs["setpoint"])
    cleanup.add(sig)
    assert sig.hints == {"fields": [sig.name]}


def test_epicssignal_sub_setpoint(cleanup, fake_motor_ioc):
    pvs = fake_motor_ioc.pvs
    pv = EpicsSignal(write_pv=pvs["setpoint"], read_pv=pvs["readback"], name="pv")
    cleanup.add(pv)

    setpoint_called = []
    setpoint_meta_called = []

    def sub_setpoint(old_value, value, **kwargs):
        setpoint_called.append((old_value, value))

    def sub_setpoint_meta(timestamp, **kwargs):
        setpoint_meta_called.append(timestamp)

    pv.subscribe(sub_setpoint, event_type=pv.SUB_SETPOINT)
    pv.subscribe(sub_setpoint_meta, event_type=pv.SUB_SETPOINT_META)

    pv.wait_for_connection()

    pv.put(1, wait=True)
    pv.put(2, wait=True)
    time.sleep(1.0)

    assert len(setpoint_called) >= 3
    assert len(setpoint_meta_called) >= 3


def test_epicssignal_get_in_callback(fake_motor_ioc, cleanup):
    pvs = fake_motor_ioc.pvs
    sig = EpicsSignal(write_pv=pvs["setpoint"], read_pv=pvs["readback"], name="motor")
    cleanup.add(sig)

    called = []

    def generic_sub(sub_type, **kwargs):
        called.append((sub_type, sig.get(), sig.get_setpoint()))

    for event_type in (
        sig.SUB_VALUE,
        sig.SUB_META,
        sig.SUB_SETPOINT,
        sig.SUB_SETPOINT_META,
    ):
        sig.subscribe(generic_sub, event_type=event_type)

    sig.wait_for_connection()

    sig.put(1, wait=True)
    sig.put(2, wait=True)
    time.sleep(0.5)

    print(called)
    # Arbitrary threshold, but if @klauer screwed something up again, this will
    # blow up
    assert len(called) < 20
    print("total", len(called))
    sig.unsubscribe_all()


@pytest.mark.motorsim
@pytest.mark.parametrize(
    "pvname, count",
    [
        ("sim:mtr1.RBV", 10),
        ("sim:mtr2.RBV", 10),
        ("sim:mtr1.RBV", 100),
        ("sim:mtr2.RBV", 100),
    ],
)
def test_epicssignal_pv_reuse(cleanup, pvname, count):
    signals = [EpicsSignal(pvname, name="sig") for i in range(count)]

    for sig in signals:
        cleanup.add(sig)
        sig.wait_for_connection()
        assert sig.connected
        assert sig.get() is not None

    if get_cl().name == "pyepics":
        assert len(set(id(sig._read_pv) for sig in signals)) == 1


@pytest.fixture(scope="function")
def path_signal(cleanup, signal_test_ioc):
    sig = EpicsPathSignal(
        signal_test_ioc.pvs["path"], name="path", path_semantics="posix"
    )
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.mark.parametrize(
    "paths",
    [
        ("C:\\some\\path\\here"),
        ("D:\\here\\is\\another\\"),
        ("C:/yet/another/path"),
        ("D:/more/paths/here/"),
    ],
)
def test_windows_paths(paths, path_signal):
    path_signal.path_semantics = "nt"
    path_signal.set(paths).wait(3)


@pytest.mark.parametrize("paths", [("/some/path/here"), ("/here/is/another/")])
def test_posix_paths(paths, path_signal):
    path_signal.set(paths).wait(3)


def test_path_semantics_exception():
    with pytest.raises(ValueError):
        EpicsPathSignal("TEST", path_semantics="not_a_thing")


def test_import_ro_signal_class():
    from ophyd import SignalRO as SignalRoFromPkg
    from ophyd.signal import SignalRO as SignalRoFromModule

    assert SignalRoFromPkg is SignalRoFromModule



# The LogFormatter is adapted light from tornado, which is licensed under
# Apache 2.0. See other_licenses/ in the repository directory.

import logging
import sys

try:
    import colorama

    colorama.init()
except ImportError:
    colorama = None
try:
    import curses
except ImportError:
    curses = None

__all__ = (
    "config_ophyd_logging",
    "get_handler",
    "logger",
    "control_layer_logger",
    "set_handler",
)


def _stderr_supports_color():
    try:
        if hasattr(sys.stderr, "isatty") and sys.stderr.isatty():
            if curses:
                curses.setupterm()
                if curses.tigetnum("colors") > 0:
                    return True
            elif colorama:
                if sys.stderr is getattr(
                    colorama.initialise, "wrapped_stderr", object()
                ):
                    return True
    except Exception:
        # Very broad exception handling because it's always better to
        # fall back to non-colored logs than to break at startup.
        pass
    return False


class LogFormatter(logging.Formatter):
    """Log formatter used in Tornado, modified for Python3-only ophyd.

    Key features of this formatter are:

    * Color support when logging to a terminal that supports it.
    * Timestamps on every log line.
    * Robust against str/bytes encoding problems.

    This formatter is enabled automatically by
    `tornado.options.parse_command_line` or `tornado.options.parse_config_file`
    (unless ``--logging=none`` is used).
    Color support on Windows versions that do not support ANSI color codes is
    enabled by use of the colorama__ library. Applications that wish to use
    this must first initialize colorama with a call to ``colorama.init``.
    See the colorama documentation for details.

    __ https://pypi.python.org/pypi/colorama

    .. versionchanged:: 4.5

       Added support for ``colorama``. Changed the constructor
       signature to be compatible with `logging.config.dictConfig`.
    """

    DEFAULT_FORMAT = "%(color)s[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d]%(end_color)s %(message)s"
    DEFAULT_DATE_FORMAT = "%y%m%d %H:%M:%S"
    DEFAULT_COLORS = {
        logging.DEBUG: 4,  # Blue
        logging.INFO: 2,  # Green
        logging.WARNING: 3,  # Yellow
        logging.ERROR: 1,  # Red
    }

    def __init__(
        self,
        fmt=DEFAULT_FORMAT,
        datefmt=DEFAULT_DATE_FORMAT,
        style="%",
        color=True,
        colors=DEFAULT_COLORS,
    ):
        r"""
        :arg bool color: Enables color support.
        :arg str fmt: Log message format.
          It will be applied to the attributes dict of log records. The
          text between ``%(color)s`` and ``%(end_color)s`` will be colored
          depending on the level if color support is on.
        :arg dict colors: color mappings from logging level to terminal color
          code
        :arg str datefmt: Datetime format.
          Used for formatting ``(asctime)`` placeholder in ``prefix_fmt``.

        .. versionchanged:: 3.2

           Added ``fmt`` and ``datefmt`` arguments.
        """
        super().__init__(datefmt=datefmt)
        self._fmt = fmt

        self._colors = {}
        if color and _stderr_supports_color():
            if curses is not None:
                # The curses module has some str/bytes confusion in
                # python3.  Until version 3.2.3, most methods return
                # bytes, but only accept strings.  In addition, we want to
                # output these strings with the logging module, which
                # works with unicode strings.  The explicit calls to
                # unicode() below are harmless in python2 but will do the
                # right conversion in python 3.
                fg_color = curses.tigetstr("setaf") or curses.tigetstr("setf") or ""

                for levelno, code in colors.items():
                    self._colors[levelno] = str(curses.tparm(fg_color, code), "ascii")
                self._normal = str(curses.tigetstr("sgr0"), "ascii")
            else:
                # If curses is not present (currently we'll only get here for
                # colorama on windows), assume hard-coded ANSI color codes.
                for levelno, code in colors.items():
                    self._colors[levelno] = "\033[2;3%dm" % code
                self._normal = "\033[0m"
        else:
            self._normal = ""

    def format(self, record):
        message = []
        if hasattr(record, "ophyd_object_name"):
            message.append(f"[{record.ophyd_object_name}]")
        elif hasattr(record, "status"):
            message.append(f"[{record.status}]")
        else:
            ...

        message.append(record.getMessage())
        record.message = " ".join(message)
        record.asctime = self.formatTime(record, self.datefmt)

        try:
            record.color = self._colors[record.levelno]
            record.end_color = self._normal
        except KeyError:
            record.color = ""
            record.end_color = ""

        formatted = self._fmt % record.__dict__

        if record.exc_info and not record.exc_text:
            record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            formatted = "{}\n{}".format(formatted.rstrip(), record.exc_text)
        return formatted.replace("\n", "\n    ")


plain_log_format = (
    "[%(levelname)1.1s %(asctime)s.%(msecs)03d %(module)s:%(lineno)d] %(message)s"
)
color_log_format = (
    "%(color)s[%(levelname)1.1s %(asctime)s.%(msecs)03d "
    "%(module)s:%(lineno)d]%(end_color)s %(message)s"
)


def validate_level(level) -> int:
    """
    Return an int for level comparison
    """
    if isinstance(level, int):
        levelno = level
    elif isinstance(level, str):
        levelno = logging.getLevelName(level)

    if isinstance(levelno, int):
        return levelno
    else:
        raise ValueError(
            "Your level is illegal, please use "
            "'CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'INFO', or 'DEBUG'."
        )


logger = logging.getLogger("ophyd")
control_layer_logger = logging.getLogger("ophyd.control_layer")


current_handler = None  # overwritten below


def config_ophyd_logging(
    file=sys.stdout, datefmt="%H:%M:%S", color=True, level="WARNING"
):
    """
    Set a new handler on the ``logging.getLogger('ophyd')`` logger.
    If this is called more than once, the handler from the previous invocation
    is removed (if still present) and replaced.

    Parameters
    ----------
    file : object with ``write`` method or filename string
        Default is ``sys.stdout``.
    datefmt : string
        Date format. Default is ``'%H:%M:%S'``.
    color : boolean
        Use ANSI color codes. True by default.
    level : str or int
        Python logging level, given as string or corresponding integer.
        Default is 'WARNING'.
    Returns
    -------
    handler : logging.Handler
        The handler, which has already been added to the 'ophyd' logger.
    Examples
    --------
    Log to a file.
    >>> config_ophyd_logging(file='/tmp/what_is_happening.txt')
    Include the date along with the time. (The log messages will always include
    microseconds, which are configured separately, not as part of 'datefmt'.)
    >>> config_ophyd_logging(datefmt="%Y-%m-%d %H:%M:%S")
    Turn off ANSI color codes.
    >>> config_ophyd_logging(color=False)
    Increase verbosity: show level DEBUG or higher.
    >>> config_ophyd_logging(level='DEBUG')
    """
    global current_handler
    if isinstance(file, str):
        handler = logging.FileHandler(file)
    else:
        handler = logging.StreamHandler(file)
    levelno = validate_level(level)
    handler.setLevel(levelno)
    if color:
        log_format = color_log_format
    else:
        log_format = plain_log_format
    handler.setFormatter(LogFormatter(log_format, datefmt=datefmt))

    if current_handler in logger.handlers:
        logger.removeHandler(current_handler)
    logger.addHandler(handler)

    current_handler = handler

    if logger.getEffectiveLevel() > levelno:
        logger.setLevel(levelno)

    return handler


set_handler = config_ophyd_logging  # for back-compat


def get_handler():
    """
    Return the handler configured by the most recent call to :func:`config_ophyd_logging`.
    If :func:`config_ophyd_logging` has not yet been called, this returns ``None``.
    """
    return current_handler



import time
import pytest

from ophyd import (
    Component as Cpt,
    SimDetector,
    SimDetectorCam,
    StatsPlugin,
    EpicsMotor,
    ROIPlugin,
    ProcessPlugin,
    Device,
)
from ophyd.areadetector.base import EpicsSignalWithRBV
from ophyd.flyers import (
    AreaDetectorTimeseriesCollector,
    WaveformCollector,
    MonitorFlyerMixin,
)
from ophyd.status import wait


@pytest.fixture(scope="function")
def ts_sim_detector(ad_prefix):
    class Detector(SimDetector):
        acquire = Cpt(EpicsSignalWithRBV, "cam1:Acquire", trigger_value=1)
        cam = Cpt(SimDetectorCam, "cam1:")
        ts_col = Cpt(AreaDetectorTimeseriesCollector, "Stats1:")
        stats = Cpt(StatsPlugin, "Stats1:")
        roi1 = Cpt(ROIPlugin, "ROI1:")
        proc1 = Cpt(ProcessPlugin, "Proc1:")

    try:
        det = Detector(ad_prefix, name="sim")
        det.wait_for_connection(timeout=1.0)
    except TimeoutError:
        pytest.skip("IOC unavailable")
    return det


@pytest.fixture
def tscollector(ts_sim_detector):
    return ts_sim_detector.ts_col


@pytest.mark.adsim
def test_ad_time_series(ts_sim_detector, tscollector):
    sim_detector = ts_sim_detector

    num_points = 3

    cam = sim_detector.cam
    cam.stage_sigs[cam.acquire_time] = 0.001
    cam.stage_sigs[cam.acquire_period] = 0.001
    cam.stage_sigs[cam.image_mode] = "Single"
    cam.stage_sigs[cam.trigger_mode] = "Internal"

    print("tscollector desc", tscollector.describe())
    print("tscollector flyer desc", tscollector.describe_collect())
    print("tscollector repr", repr(tscollector))
    print("simdet stage sigs", sim_detector.stage_sigs)
    print("tscoll stage sigs", tscollector.stage_sigs)
    print("cam stage sigs", cam.stage_sigs)
    print("stats stage sigs", sim_detector.stats.stage_sigs)

    try:
        # In case the collector is currently running...
        wait(tscollector.complete())
    except RuntimeError:
        ...

    tscollector.stage_sigs[tscollector.num_points] = num_points

    sim_detector.stage()
    tscollector.kickoff()

    for i in range(num_points):
        st = sim_detector.trigger()
        wait(st)
        print(st)
        time.sleep(0.1)

    tscollector.pause()
    tscollector.resume()

    st = tscollector.complete()
    wait(st)

    collected = list(tscollector.collect())
    print("collected", collected)
    sim_detector.unstage()
    assert len(collected) == num_points
    # TODO any more validation here?


@pytest.fixture(scope="function")
def wf_sim_detector(ad_prefix):
    suffix = "??TODO??"

    class Detector(SimDetector):
        wfcol = Cpt(WaveformCollector, suffix)

    try:
        det = Detector(ad_prefix, name="det")
        det.wait_for_connection(timeout=1.0)
    except TimeoutError:
        pytest.skip("IOC unavailable")
    return det


@pytest.fixture
def wfcol(wf_sim_detector):
    return wf_sim_detector.wfcol


@pytest.mark.adsim
def test_waveform(wf_sim_detector, wfcol):
    print("waveform collector", wfcol)
    print("wfcol flyer desc", wfcol.describe_collect())


@pytest.mark.motorsim
@pytest.mark.parametrize("pivot", [True, False])
def test_monitor_flyer(pivot):
    class BasicDevice(Device):
        mtr1 = Cpt(EpicsMotor, "XF:31IDA-OP{Tbl-Ax:X2}Mtr")
        mtr2 = Cpt(EpicsMotor, "XF:31IDA-OP{Tbl-Ax:X3}Mtr")

    class FlyerDevice(MonitorFlyerMixin, BasicDevice):
        pass

    fdev = FlyerDevice(
        "", name="fdev", stream_names={"mtr1.user_readback": "oranges"}, pivot=pivot
    )
    fdev.wait_for_connection()

    fdev.monitor_attrs = ["mtr1.user_readback", "mtr2.user_readback"]
    fdev.describe()

    st = fdev.kickoff()
    wait(st)

    mtr1, mtr2 = fdev.mtr1, fdev.mtr2
    rbv1, rbv2 = mtr1.position, mtr2.position
    fdev.mtr1.move(rbv1 + 0.2, wait=True)
    fdev.mtr2.move(rbv2 + 0.2, wait=True)

    fdev.pause()

    fdev.mtr1.move(rbv1 - 0.2, wait=True)
    fdev.mtr2.move(rbv2 - 0.2, wait=True)

    fdev.resume()
    st = fdev.complete()
    wait(st)

    print(fdev.describe_collect())

    desc1 = fdev.mtr1.user_readback.describe()
    desc2 = fdev.mtr2.user_readback.describe()

    if not pivot:
        desc1[fdev.mtr1.user_readback.name]["dtype"] = "array"
        desc2[fdev.mtr2.user_readback.name]["dtype"] = "array"

    assert fdev.describe_collect() == {
        "oranges": desc1,
        "fdev_mtr2": desc2,
    }
    data = list(fdev.collect())
    print("collected data", data)
    # data from both motors

    if not pivot:
        assert len(data) == 2
        d1 = data[0]["data"]["fdev_mtr1"]
        d2 = data[1]["data"]["fdev_mtr2"]

    else:
        assert len(data) >= 2

        d1 = [d["data"]["fdev_mtr1"] for d in data if "fdev_mtr1" in d["data"]]
        d2 = [d["data"]["fdev_mtr2"] for d in data if "fdev_mtr2" in d["data"]]

    # and at least more than one data point...
    assert len(d1) > 1
    assert len(d2) > 1

    print("data1", d1)
    print("data2", d2)
    # raise ValueError()



import inspect
import os.path
import textwrap
from typing import Dict, List, Type

import ophyd

OPHYD_SKIP = {
    # Methods
    # 'check_value',
    "clear_sub",
    # 'configure',
    # 'describe',
    # 'describe_configuration',
    "destroy",
    # 'get',
    "get_device_tuple",
    "get_instantiated_signals",
    "pause",
    "put",  # prefer `set`
    # 'read',
    # 'read_configuration',
    "resume",
    # 'stage',
    # 'stop',
    # 'subscribe',
    # 'summary',
    # 'trigger',
    # 'unstage',
    "unsubscribe",
    "unsubscribe_all",
    "wait_for_connection",
    "walk_components",
    "walk_signals",
    "walk_subdevice_classes",
    "walk_subdevices",
    # Attributes
    "SUB_ACQ_DONE",
    "SUB_DONE",
    "SUB_READBACK",
    "SUB_START",
    "SUB_VALUE",
    "attr_name",
    "component_names",
    # 'configuration_attrs',
    # 'connected',
    "dotted_name",
    "event_types",
    # 'hints',
    # 'kind',
    "lazy_wait_for_connection",
    # 'lightpath_cpts',
    "name",
    "parent",
    "read_attrs",
    "removed",
    "report",
    "root",
    "signal_names",
    "trigger_signals",
}


short_component_names = {
    ophyd.Component: "",
    ophyd.DynamicDeviceComponent: "DDC",
    ophyd.FormattedComponent: "FCpt",
    # NOTE: Users may add in their own component types here, without
    # re-implementing these documentation helpers
}


def get_class_info(cls: type) -> Dict[str, str]:
    """
    Get class information in a jinja-friendly way.

    Keys for usage include ``name`` (the class name), ``full_name`` (a
    fully-qualified module and class name), ``class`` (the class type itself),
    ``link`` (a link to the class, only showing its name), and ``full_link`` (a
    full link showing the fully-qualified name).

    Parameters
    ----------
    cls : type
        Class type.
    """
    if cls is None:
        return None

    def link_as(text):
        return f":class:`{text} <{full_name}>`"

    full_name = f"{cls.__module__}.{cls.__name__}"
    return {
        "name": cls.__name__,
        "full_name": full_name,
        "class": cls,
        "full_link": f":class:`{full_name}`",
        "link": f":class:`~{full_name}`",
        "link_as": link_as,
    }


def _dynamic_device_component_to_row(base_attrs, cls, attr, cpt):
    """
    Convert an :class:`ophyd.DynamicDeviceComponent` to a table row dictionary.

    Parameters
    ----------
    base_attrs : dict
        Dictionary indicating where base attributes came from.

    cls : ophyd.Device subclass
        The parent class.

    attr : str
        The component attribute name.

    cpt : ophyd.DynamicDeviceComponent
        The component.
    """
    cpt_type = short_component_names.get(type(cpt), type(cpt).__name__)

    doc = cpt.doc or ""
    if doc.startswith(f"{cpt.__class__.__name__} attribute"):
        doc = ""

    nested_components = [
        _component_to_row(base_attrs, cls, attr, dynamic_cpt)
        for attr, dynamic_cpt in cpt.components.items()
    ]

    return dict(
        component=cpt,
        attr=attr if not cpt_type else f"{attr} ({cpt_type})",
        cls=get_class_info(getattr(cpt, "cls", None)),
        nested_components=nested_components,
        doc=doc,
        kind=cpt.kind.name,
        inherited_from=get_class_info(base_attrs.get(attr, None)),
    )


def _component_to_row(base_attrs, cls, attr, cpt):
    """
    Convert an :class:`ophyd.Component` to a table row dictionary.

    Parameters
    ----------
    base_attrs : dict
        Dictionary indicating where base attributes came from.

    cls : ophyd.Device subclass
        The parent class.

    attr : str
        The component attribute name.

    cpt : ophyd.Component
        The component.
    """
    if isinstance(cpt, ophyd.DynamicDeviceComponent):
        return _dynamic_device_component_to_row(base_attrs, cls, attr, cpt)

    cpt_type = short_component_names.get(type(cpt), type(cpt).__name__)

    doc = cpt.doc or ""
    if doc.startswith(f"{cpt.__class__.__name__} attribute"):
        doc = ""

    if doc.startswith("AreaDetector Component") and "::" in doc:
        _, doc = doc.split("::", 1)
        doc = textwrap.dedent(doc).strip()

    return dict(
        component=cpt,  # access to the component instance itself
        attr=attr if not cpt_type else f"{attr} ({cpt_type})",
        cls=get_class_info(getattr(cpt, "cls", None)),
        suffix=f"``{cpt.suffix}``" if cpt.suffix else "",
        doc=doc,
        kind=cpt.kind.name,
        inherited_from=get_class_info(base_attrs.get(attr, None)),
    )


def _get_base_attrs(cls: Type[ophyd.Device]) -> Dict[str, Type[ophyd.Device]]:
    """
    Determine which components came from which base class.

    Parameters
    ----------
    cls : ophyd.Device subclass
    """
    base_devices = [
        base for base in reversed(cls.__bases__) if hasattr(base, "_sig_attrs")
    ]

    return {
        attr: base for base in base_devices for attr, cpt in base._sig_attrs.items()
    }


# NOTE: can't use functools.lru_cache here as it's not picklable
# (or at least this was the case when embedded in conf.py)
_device_cache = {}


def _get_device_info(cls: Type[ophyd.Device]) -> List[Dict]:
    """
    Get Device information that can easily be rendered as a table.

    Parameters
    ----------
    cls : ophyd.Device subclass
    """
    if not issubclass(cls, ophyd.Device):
        return []

    base_attrs = _get_base_attrs(cls)

    return [
        _component_to_row(base_attrs, cls, attr, cpt)
        for attr, cpt in cls._sig_attrs.items()
    ]


def get_device_info(module, name):
    """
    Get Device information that can easily be rendered as a table.

    This is the hook the jinja ``class.rst`` templates use.

    Parameters
    ----------
    module : str
        Module name.

    name : str
        Class name.
    """
    class_name = f"{module}.{name}"
    if class_name not in _device_cache:
        module_name, class_name = class_name.rsplit(".", 1)
        module = __import__(module_name, globals(), locals(), [class_name])
        cls = getattr(module, class_name)
        _device_cache[class_name] = _get_device_info(cls)
    return _device_cache[class_name]


def skip_components_and_ophyd_stuff(app, what, name, obj, skip, options):
    """
    Skip components, leaving only the table information.

    Also, filter out unimportant attributes based on ``OPHYD_SKIP``.
    """
    if isinstance(obj, ophyd.Component):
        return True

    if name.startswith("_"):
        # It's unclear if I broke this or if it's always been broken,
        # but for our use case we never want to document `_` items with
        # autoclass.
        return True

    if name in OPHYD_SKIP:
        return True

    return skip


def rst_with_jinja(app, docname, source):
    """
    Render our pages as a jinja template for fancy templating goodness.

    Usage
    -----

    .. code::

        def setup(app):
            app.connect("source-read", rst_with_jinja)
    """
    # Borrowed from
    # https://www.ericholscher.com/blog/2016/jul/25/integrating-jinja-rst-sphinx/

    # Make sure we're outputting HTML
    if app.builder.format == "html":
        rendered = app.builder.templates.render_string(
            source[0], app.config.html_context
        )
        source[0] = rendered


def setup(app):
    """
    Setup method to configure caproto sphinx documentation helpers.

    This is a convenience method; advanced users may wish to replicate this
    method on their own.

    Parameters
    ----------
    app : sphinx.Application

    """
    app.connect("autodoc-skip-member", skip_components_and_ophyd_stuff)
    app.connect("source-read", rst_with_jinja)


# The following are convenient defaults that users may import in their
# conf.py:
autosummary_context = {
    # Allow autosummary/class.rst to do its magic:
    "get_device_info": get_device_info,
    "inspect": inspect,
    "path": os.path,
    # Where is your project root, relative to conf.py?
    "project_root": "..",
    # To change where generated source files go, change the following in
    # conf.py:
    "generated_toctree": "generated",
}

autodoc_default_options = {
    "members": "",
    "member-order": "bysource",
    "special-members": "",
    "undoc-members": False,
    "exclude-members": "",
}
autoclass_content = "init"  # otherwise duplicates will be generated

intersphinx_mapping = {
    "ophyd": ("https://blueskyproject.io/ophyd", None),
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://docs.scipy.org/doc/numpy", None),
}



import pytest
import os
import logging
import numpy as np
import tempfile

from ophyd.utils import epics_pvs as epics_utils
from ophyd.utils import make_dir_tree, makedirs
from ophyd import Signal


logger = logging.getLogger(__name__)


def test_split():
    utils = epics_utils

    assert utils.split_record_field("record.field") == ("record", "field")
    assert utils.split_record_field("record.field.invalid") == (
        "record.field",
        "invalid",
    )
    assert utils.strip_field("record.field") == "record"
    assert utils.strip_field("record.field.invalid") == "record.field"
    assert utils.record_field("record", "field") == "record.FIELD"


def test_waveform_to_string():
    s = "abcdefg"
    asc = [ord(c) for c in s]
    assert epics_utils.waveform_to_string(asc) == s

    asc = [ord(c) for c in s] + [0, 0, 0]
    assert epics_utils.waveform_to_string(asc) == s


def test_records_from_db():
    # db_dir = os.path.join(config.epics_base, 'db')

    # if os.path.exists(db_dir):
    #     # fall back on the db file included with the tests
    db_dir = os.path.dirname(__file__)
    db_path = os.path.join(db_dir, "scaler.db")
    records = epics_utils.records_from_db(db_path)
    assert ("bo", "$(P)$(S)_calcEnable") in records


@pytest.mark.parametrize(
    "value, dtype, shape",
    [
        [1, "integer", []],
        [1.0, "number", []],
        [1e-3, "number", []],
        ["foo", "string", []],
        [np.array([1, 2, 3]), "array", [3]],
        [np.array([[1, 2], [3, 4]]), "array", [2, 2]],
        [(1, 2, 3), "array", [3]],
        [[1, 2, 3], "array", [3]],
        [[], "array", [0]],
    ],
)
def test_data_type_and_shape(value, dtype, shape):
    utils = epics_utils
    assert utils.data_type(value) == dtype
    assert utils.data_shape(value) == shape


@pytest.mark.parametrize("value", [dict()])
def test_invalid_data_type(value):
    utils = epics_utils
    with pytest.raises(ValueError):
        utils.data_type(value)


def assert_OD_equal_ignore_ts(a, b):
    for (k1, v1), (k2, v2) in zip(a.items(), b.items()):
        assert (k1 == k2) and (v1["value"] == v2["value"])


def assert_file_mode(path, expected):
    assert (os.stat(path).st_mode & 0o777) == expected


def test_makedirs():
    with tempfile.TemporaryDirectory() as tempdir:
        create_dir = os.path.join(tempdir, "a")
        makedirs(create_dir, mode=0o767, mode_base=tempdir)
        assert_file_mode(create_dir, 0o767)


def test_make_dir_tree():
    with tempfile.TemporaryDirectory() as tempdir:
        paths = make_dir_tree(2016, base_path=tempdir, mode=0o777)
        assert len(paths) == 366

        for path in paths:
            assert_file_mode(path, 0o777)

        assert os.path.join(tempdir, "2016", "03", "04") in paths
        assert os.path.join(tempdir, "2016", "02", "29") in paths


def test_valid_pvname():
    with pytest.raises(epics_utils.BadPVName):
        epics_utils.validate_pv_name("this.will.fail")


def test_array_into_softsignal():
    data = np.array([1, 2, 3])
    s = Signal(name="np.array")
    s.set(data).wait()
    assert np.all(s.get() == data)


def test_none_signal():
    import itertools

    class CycleSignal(Signal):
        def __init__(self, *args, value_cycle, **kwargs):
            super().__init__(*args, **kwargs)
            self._value_cycle = itertools.cycle(value_cycle)

        def get(self):
            return next(self._value_cycle)

    cs = CycleSignal(
        name="cycle", value_cycle=[0, 1, 2, None, 4], tolerance=0.01, rtolerance=0.1
    )

    cs.set(4).wait()


def test_set_signal_to_None():
    s = Signal(value="0", name="bob")
    s.set(None).wait(timeout=1)
